import pytest
import json
from flask import Flask
from unittest.mock import patch
from mlflow_oidc_auth.responses.client_error import (
    make_auth_required_response,
    make_forbidden_response,
    make_basic_auth_response,
)


@pytest.fixture(scope="module")
def test_app():
    app = Flask(__name__)
    with app.app_context():
        yield app


class TestClientErrorResponses:
    def test_make_auth_required_response(self, test_app):
        response = make_auth_required_response()
        assert response.status_code == 401
        assert response.get_json() == {"message": "Authentication required"}

    def test_make_forbidden_response(self, test_app):
        response = make_forbidden_response()
        assert response.status_code == 403
        assert response.get_json() == {"message": "Permission denied"}

    def test_make_forbidden_response_custom_message(self, test_app):
        custom_msg = {"message": "Custom permission denied message"}
        response = make_forbidden_response(custom_msg)
        assert response.status_code == 403
        assert response.get_json() == custom_msg

    def test_make_basic_auth_response(self, test_app):
        response = make_basic_auth_response()
        assert response.status_code == 401
        assert response.data.decode() == ("You are not authenticated. Please see documentation for details" "https://github.com/mlflow-oidc/mlflow-oidc-auth")
        assert response.headers["WWW-Authenticate"] == 'Basic realm="mlflow"'


class TestClientErrorResponseSecurity:
    """Test security aspects of client error responses."""

    def test_auth_required_response_security_headers(self, test_app):
        """Test that auth required response has appropriate security characteristics."""
        response = make_auth_required_response()

        # Verify response is JSON and properly formatted
        assert response.content_type == "application/json"
        assert response.status_code == 401

        # Verify response doesn't leak sensitive information
        response_data = response.get_json()
        assert "message" in response_data
        assert len(response_data) == 1  # Only contains expected message

        # Verify message is safe and doesn't contain sensitive data
        assert "Authentication required" in response_data["message"]
        assert "password" not in response_data["message"].lower()
        assert "token" not in response_data["message"].lower()
        assert "secret" not in response_data["message"].lower()

    def test_forbidden_response_security_headers(self, test_app):
        """Test that forbidden response has appropriate security characteristics."""
        response = make_forbidden_response()

        # Verify response is JSON and properly formatted
        assert response.content_type == "application/json"
        assert response.status_code == 403

        # Verify response doesn't leak sensitive information
        response_data = response.get_json()
        assert "message" in response_data
        assert len(response_data) == 1  # Only contains expected message

        # Verify message is safe and doesn't contain sensitive data
        assert "Permission denied" in response_data["message"]
        assert "password" not in response_data["message"].lower()
        assert "token" not in response_data["message"].lower()
        assert "secret" not in response_data["message"].lower()

    def test_basic_auth_response_security_headers(self, test_app):
        """Test that basic auth response has appropriate security characteristics."""
        response = make_basic_auth_response()

        # Verify response has proper WWW-Authenticate header
        assert "WWW-Authenticate" in response.headers
        assert response.headers["WWW-Authenticate"] == 'Basic realm="mlflow"'

        # Verify response doesn't leak sensitive information
        response_text = response.data.decode()
        assert "password" not in response_text.lower()
        assert "token" not in response_text.lower()
        assert "secret" not in response_text.lower()

        # Verify it contains helpful documentation link
        assert "https://github.com/mlflow-oidc/mlflow-oidc-auth" in response_text

    def test_forbidden_response_custom_message_sanitization(self, test_app):
        """Test that custom messages in forbidden responses are properly handled."""
        # Test with potentially dangerous custom message
        dangerous_msg = {"message": "Access denied", "debug_info": "Internal server details", "user_id": "12345"}

        response = make_forbidden_response(dangerous_msg)
        assert response.status_code == 403

        # Verify the entire custom message is preserved (responsibility of caller to sanitize)
        response_data = response.get_json()
        assert response_data == dangerous_msg

    def test_forbidden_response_with_none_message(self, test_app):
        """Test forbidden response when explicitly passed None."""
        response = make_forbidden_response(None)
        assert response.status_code == 403
        assert response.get_json() == {"message": "Permission denied"}

    def test_forbidden_response_with_empty_dict(self, test_app):
        """Test forbidden response with empty dictionary."""
        response = make_forbidden_response({})
        assert response.status_code == 403
        assert response.get_json() == {}


class TestClientErrorResponseConsistency:
    """Test consistency and user experience aspects of error responses."""

    def test_response_format_consistency(self, test_app):
        """Test that all JSON responses follow consistent format."""
        auth_response = make_auth_required_response()
        forbidden_response = make_forbidden_response()

        # Both should be JSON responses
        assert auth_response.content_type == "application/json"
        assert forbidden_response.content_type == "application/json"

        # Both should have message field in default case
        auth_data = auth_response.get_json()
        forbidden_data = forbidden_response.get_json()

        assert "message" in auth_data
        assert "message" in forbidden_data
        assert isinstance(auth_data["message"], str)
        assert isinstance(forbidden_data["message"], str)

    def test_status_code_consistency(self, test_app):
        """Test that status codes are consistent with HTTP standards."""
        auth_response = make_auth_required_response()
        forbidden_response = make_forbidden_response()
        basic_auth_response = make_basic_auth_response()

        # Verify correct HTTP status codes
        assert auth_response.status_code == 401  # Unauthorized
        assert forbidden_response.status_code == 403  # Forbidden
        assert basic_auth_response.status_code == 401  # Unauthorized (with WWW-Authenticate)

    def test_error_message_clarity(self, test_app):
        """Test that error messages are clear and helpful."""
        auth_response = make_auth_required_response()
        forbidden_response = make_forbidden_response()
        basic_auth_response = make_basic_auth_response()

        # Verify messages are clear and actionable
        auth_data = auth_response.get_json()
        forbidden_data = forbidden_response.get_json()
        basic_text = basic_auth_response.data.decode()

        assert "Authentication required" in auth_data["message"]
        assert "Permission denied" in forbidden_data["message"]
        assert "not authenticated" in basic_text
        assert "documentation" in basic_text


class TestClientErrorResponseSerialization:
    """Test response serialization and content negotiation."""

    def test_json_serialization_auth_required(self, test_app):
        """Test JSON serialization for auth required response."""
        response = make_auth_required_response()

        # Verify response can be serialized to JSON
        json_data = response.get_json()
        assert json_data is not None

        # Verify JSON is valid by re-serializing
        json_string = json.dumps(json_data)
        assert json_string is not None

        # Verify deserialization works
        deserialized = json.loads(json_string)
        assert deserialized == json_data

    def test_json_serialization_forbidden(self, test_app):
        """Test JSON serialization for forbidden response."""
        response = make_forbidden_response()

        # Verify response can be serialized to JSON
        json_data = response.get_json()
        assert json_data is not None

        # Verify JSON is valid by re-serializing
        json_string = json.dumps(json_data)
        assert json_string is not None

        # Verify deserialization works
        deserialized = json.loads(json_string)
        assert deserialized == json_data

    def test_json_serialization_custom_message(self, test_app):
        """Test JSON serialization with custom message."""
        custom_msg = {"message": "Custom error", "code": "ERR_001", "details": {"field": "value"}}

        response = make_forbidden_response(custom_msg)

        # Verify response can be serialized to JSON
        json_data = response.get_json()
        assert json_data is not None

        # Verify JSON is valid by re-serializing
        json_string = json.dumps(json_data)
        assert json_string is not None

        # Verify deserialization works and preserves structure
        deserialized = json.loads(json_string)
        assert deserialized == custom_msg

    def test_content_type_headers(self, test_app):
        """Test that content type headers are set correctly."""
        auth_response = make_auth_required_response()
        forbidden_response = make_forbidden_response()
        basic_auth_response = make_basic_auth_response()

        # JSON responses should have application/json content type
        assert auth_response.content_type == "application/json"
        assert forbidden_response.content_type == "application/json"

        # Basic auth response should have text/html content type (Flask default for string)
        assert basic_auth_response.content_type == "text/html; charset=utf-8"

    def test_response_encoding(self, test_app):
        """Test that responses are properly encoded."""
        auth_response = make_auth_required_response()
        forbidden_response = make_forbidden_response()
        basic_auth_response = make_basic_auth_response()

        # Verify responses can be decoded without errors
        auth_data = auth_response.data.decode("utf-8")
        forbidden_data = forbidden_response.data.decode("utf-8")
        basic_data = basic_auth_response.data.decode("utf-8")

        assert auth_data is not None
        assert forbidden_data is not None
        assert basic_data is not None

        # Verify JSON responses contain valid JSON
        json.loads(auth_data)  # Should not raise exception
        json.loads(forbidden_data)  # Should not raise exception


class TestClientErrorResponseEdgeCases:
    """Test edge cases and error conditions."""

    def test_forbidden_response_with_non_dict_message(self, test_app):
        """Test forbidden response with non-dictionary message."""
        # Test with string message
        response = make_forbidden_response("String message")
        assert response.status_code == 403
        assert response.get_json() == "String message"

        # Test with list message
        response = make_forbidden_response(["item1", "item2"])
        assert response.status_code == 403
        assert response.get_json() == ["item1", "item2"]

        # Test with number message
        response = make_forbidden_response(42)
        assert response.status_code == 403
        assert response.get_json() == 42

    def test_response_immutability(self, test_app):
        """Test that responses are properly constructed and immutable."""
        response1 = make_auth_required_response()
        response2 = make_auth_required_response()

        # Responses should be independent instances
        assert response1 is not response2
        assert response1.status_code == response2.status_code
        assert response1.get_json() == response2.get_json()

    @patch("mlflow_oidc_auth.responses.client_error.make_response")
    def test_make_response_error_handling(self, mock_make_response, test_app):
        """Test error handling when make_response fails."""
        mock_make_response.side_effect = Exception("Flask error")

        with pytest.raises(Exception, match="Flask error"):
            make_auth_required_response()

    @patch("mlflow_oidc_auth.responses.client_error.jsonify")
    def test_jsonify_error_handling(self, mock_jsonify, test_app):
        """Test error handling when jsonify fails."""
        mock_jsonify.side_effect = Exception("JSON error")

        with pytest.raises(Exception, match="JSON error"):
            make_auth_required_response()

    def test_large_custom_message_handling(self, test_app):
        """Test handling of large custom messages."""
        large_message = {"message": "A" * 10000, "details": "B" * 5000}  # Large message

        response = make_forbidden_response(large_message)
        assert response.status_code == 403

        # Verify large message is handled correctly
        response_data = response.get_json()
        assert response_data["message"] == "A" * 10000
        assert response_data["details"] == "B" * 5000



"""token expiration

Revision ID: 0565cf04c12e
Revises: 1ddf23934b63
Create Date: 2025-05-22 23:41:23.392286

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0565cf04c12e"
down_revision = "1ddf23934b63"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "users",
        sa.Column("password_expiration", sa.DateTime(), nullable=True),
    )


def downgrade() -> None:
    op.drop_column(
        "users",
        "password_expiration",
    )



from datetime import datetime

from sqlalchemy import Boolean, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from mlflow_oidc_auth.entities import (
    ExperimentGroupRegexPermission,
    ExperimentPermission,
    ExperimentRegexPermission,
    Group,
    RegisteredModelGroupRegexPermission,
    RegisteredModelPermission,
    RegisteredModelRegexPermission,
    ScorerGroupPermission,
    ScorerGroupRegexPermission,
    ScorerPermission,
    ScorerRegexPermission,
    User,
    UserGroup,
)


class Base(DeclarativeBase):
    pass


class SqlUser(Base):
    __tablename__ = "users"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    username: Mapped[str] = mapped_column(String(255), unique=True)
    display_name: Mapped[str] = mapped_column(String(255), nullable=False)
    password_hash: Mapped[str] = mapped_column(String(255), nullable=False)
    password_expiration: Mapped[datetime] = mapped_column(nullable=True)
    is_admin: Mapped[bool] = mapped_column(Boolean, default=False)
    is_service_account: Mapped[bool] = mapped_column(Boolean, default=False)
    experiment_permissions: Mapped[list["SqlExperimentPermission"]] = relationship("SqlExperimentPermission", backref="users")
    registered_model_permissions: Mapped[list["SqlRegisteredModelPermission"]] = relationship("SqlRegisteredModelPermission", backref="users")
    scorer_permissions: Mapped[list["SqlScorerPermission"]] = relationship("SqlScorerPermission", backref="users")
    groups: Mapped[list["SqlGroup"]] = relationship(
        "SqlGroup",
        secondary="user_groups",
        back_populates="users",
    )

    def to_mlflow_entity(self):
        return User(
            id_=self.id,
            username=self.username,
            display_name=self.display_name,
            password_hash=self.password_hash,
            password_expiration=self.password_expiration,
            is_admin=self.is_admin,
            is_service_account=self.is_service_account,
            experiment_permissions=[p.to_mlflow_entity() for p in self.experiment_permissions],
            registered_model_permissions=[p.to_mlflow_entity() for p in self.registered_model_permissions],
            scorer_permissions=[p.to_mlflow_entity() for p in self.scorer_permissions],
            groups=[g.to_mlflow_entity() for g in self.groups],
        )


class SqlExperimentPermission(Base):
    __tablename__ = "experiment_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    experiment_id: Mapped[str] = mapped_column(String(255), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("experiment_id", "user_id", name="unique_experiment_user"),)

    def to_mlflow_entity(self):
        return ExperimentPermission(
            experiment_id=self.experiment_id,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlRegisteredModelPermission(Base):
    __tablename__ = "registered_model_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("name", "user_id", name="unique_name_user"),)

    def to_mlflow_entity(self):
        return RegisteredModelPermission(
            name=self.name,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlGroup(Base):
    __tablename__ = "groups"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    group_name: Mapped[str] = mapped_column(String(255), nullable=False)
    __table_args__ = (UniqueConstraint("group_name"),)
    users: Mapped[list["SqlUser"]] = relationship(
        "SqlUser",
        secondary="user_groups",
        back_populates="groups",
    )

    def to_mlflow_entity(self):
        return Group(
            id_=self.id,
            group_name=self.group_name,
        )


class SqlUserGroup(Base):
    __tablename__ = "user_groups"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    __table_args__ = (UniqueConstraint("user_id", "group_id", name="unique_user_group"),)

    def to_mlflow_entity(self):
        return UserGroup(
            user_id=self.user_id,
            group_id=self.group_id,
        )


class SqlExperimentGroupPermission(Base):
    __tablename__ = "experiment_group_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    experiment_id: Mapped[str] = mapped_column(String(255), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("experiment_id", "group_id", name="unique_experiment_group"),)

    def to_mlflow_entity(self):
        return ExperimentPermission(
            experiment_id=self.experiment_id,
            group_id=self.group_id,
            permission=self.permission,
        )


class SqlRegisteredModelGroupPermission(Base):
    __tablename__ = "registered_model_group_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    prompt: Mapped[bool] = mapped_column(Boolean, default=False)
    __table_args__ = (UniqueConstraint("name", "group_id", name="unique_name_group"),)

    def to_mlflow_entity(self):
        return RegisteredModelPermission(
            name=self.name,
            group_id=self.group_id,
            permission=self.permission,
            prompt=bool(self.prompt),
        )


class SqlExperimentRegexPermission(Base):
    __tablename__ = "experiment_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("regex", "user_id", name="unique_experiment_user_regex"),)

    def to_mlflow_entity(self):
        return ExperimentRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlRegisteredModelRegexPermission(Base):
    __tablename__ = "registered_model_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    prompt: Mapped[bool] = mapped_column(Boolean, default=False)
    __table_args__ = (UniqueConstraint("regex", "user_id", "prompt", name="unique_name_user_regex"),)

    def to_mlflow_entity(self):
        return RegisteredModelRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            user_id=self.user_id,
            permission=self.permission,
            prompt=bool(self.prompt),
        )


class SqlExperimentGroupRegexPermission(Base):
    __tablename__ = "experiment_group_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("regex", "group_id", name="unique_experiment_group_regex"),)

    def to_mlflow_entity(self):
        return ExperimentGroupRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            group_id=self.group_id,
            permission=self.permission,
        )


class SqlRegisteredModelGroupRegexPermission(Base):
    __tablename__ = "registered_model_group_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    prompt: Mapped[bool] = mapped_column(Boolean, default=False)
    __table_args__ = (UniqueConstraint("regex", "group_id", "prompt", name="unique_name_group_regex"),)

    def to_mlflow_entity(self):
        return RegisteredModelGroupRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            group_id=self.group_id,
            permission=self.permission,
            prompt=bool(self.prompt),
        )


class SqlScorerPermission(Base):
    __tablename__ = "scorer_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    experiment_id: Mapped[str] = mapped_column(String(255), nullable=False)
    scorer_name: Mapped[str] = mapped_column(String(256), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("experiment_id", "scorer_name", "user_id", name="unique_scorer_user"),)

    def to_mlflow_entity(self):
        return ScorerPermission(
            experiment_id=self.experiment_id,
            scorer_name=self.scorer_name,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlScorerGroupPermission(Base):
    __tablename__ = "scorer_group_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    experiment_id: Mapped[str] = mapped_column(String(255), nullable=False)
    scorer_name: Mapped[str] = mapped_column(String(256), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("experiment_id", "scorer_name", "group_id", name="unique_scorer_group"),)

    def to_mlflow_entity(self):
        return ScorerGroupPermission(
            experiment_id=self.experiment_id,
            scorer_name=self.scorer_name,
            group_id=self.group_id,
            permission=self.permission,
        )


class SqlScorerRegexPermission(Base):
    __tablename__ = "scorer_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("regex", "user_id", name="unique_scorer_user_regex"),)

    def to_mlflow_entity(self):
        return ScorerRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlScorerGroupRegexPermission(Base):
    __tablename__ = "scorer_group_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("regex", "group_id", name="unique_scorer_group_regex"),)

    def to_mlflow_entity(self):
        return ScorerGroupRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            group_id=self.group_id,
            permission=self.permission,
        )



from datetime import datetime
from typing import List, Optional

from mlflow.store.db.utils import _get_managed_session_maker, create_sqlalchemy_engine_with_retry
from mlflow.utils.uri import extract_db_type_from_uri
from sqlalchemy.orm import sessionmaker

from mlflow_oidc_auth.db import utils as dbutils
from mlflow_oidc_auth.entities import (
    ExperimentGroupRegexPermission,
    ExperimentPermission,
    ExperimentRegexPermission,
    RegisteredModelGroupRegexPermission,
    RegisteredModelPermission,
    RegisteredModelRegexPermission,
    ScorerGroupRegexPermission,
    ScorerPermission,
    ScorerRegexPermission,
    User,
)
from mlflow_oidc_auth.repository import (
    ExperimentPermissionGroupRegexRepository,
    ExperimentPermissionGroupRepository,
    ExperimentPermissionRegexRepository,
    ExperimentPermissionRepository,
    GroupRepository,
    PromptPermissionGroupRepository,
    RegisteredModelGroupRegexPermissionRepository,
    RegisteredModelPermissionGroupRepository,
    RegisteredModelPermissionRegexRepository,
    RegisteredModelPermissionRepository,
    ScorerPermissionGroupRegexRepository,
    ScorerPermissionGroupRepository,
    ScorerPermissionRegexRepository,
    ScorerPermissionRepository,
    UserRepository,
)


class SqlAlchemyStore:
    def init_db(self, db_uri):
        self.db_uri = db_uri
        self.db_type = extract_db_type_from_uri(db_uri)
        self.engine = create_sqlalchemy_engine_with_retry(db_uri)
        dbutils.migrate_if_needed(self.engine, "head")
        SessionMaker = sessionmaker(bind=self.engine)
        self.ManagedSessionMaker = _get_managed_session_maker(SessionMaker, self.db_type)
        self.user_repo = UserRepository(self.ManagedSessionMaker)
        self.experiment_repo = ExperimentPermissionRepository(self.ManagedSessionMaker)
        self.experiment_group_repo = ExperimentPermissionGroupRepository(self.ManagedSessionMaker)
        self.group_repo = GroupRepository(self.ManagedSessionMaker)
        self.registered_model_repo = RegisteredModelPermissionRepository(self.ManagedSessionMaker)
        self.registered_model_group_repo = RegisteredModelPermissionGroupRepository(self.ManagedSessionMaker)
        self.prompt_group_repo = PromptPermissionGroupRepository(self.ManagedSessionMaker)
        self.experiment_regex_repo = ExperimentPermissionRegexRepository(self.ManagedSessionMaker)
        self.experiment_group_regex_repo = ExperimentPermissionGroupRegexRepository(self.ManagedSessionMaker)
        self.registered_model_regex_repo = RegisteredModelPermissionRegexRepository(self.ManagedSessionMaker)
        self.registered_model_group_regex_repo = RegisteredModelGroupRegexPermissionRepository(self.ManagedSessionMaker)
        self.prompt_group_regex_repo = RegisteredModelGroupRegexPermissionRepository(self.ManagedSessionMaker)
        self.prompt_regex_repo = RegisteredModelPermissionRegexRepository(self.ManagedSessionMaker)

        # Scorer permissions
        self.scorer_repo = ScorerPermissionRepository(self.ManagedSessionMaker)
        self.scorer_group_repo = ScorerPermissionGroupRepository(self.ManagedSessionMaker)
        self.scorer_regex_repo = ScorerPermissionRegexRepository(self.ManagedSessionMaker)
        self.scorer_group_regex_repo = ScorerPermissionGroupRegexRepository(self.ManagedSessionMaker)

    def ping(self) -> bool:
        """Lightweight database connectivity check for health probes.

        Returns:
            True if database is reachable, False otherwise.
        """
        from sqlalchemy import text

        try:
            with self.engine.connect() as connection:
                connection.execute(text("SELECT 1"))
            return True
        except Exception:
            return False

    # Scorer CRUD (user-scoped)
    def create_scorer_permission(self, experiment_id: str, scorer_name: str, username: str, permission: str) -> ScorerPermission:
        return self.scorer_repo.grant_permission(experiment_id, scorer_name, username, permission)

    def get_scorer_permission(self, experiment_id: str, scorer_name: str, username: str) -> ScorerPermission:
        return self.scorer_repo.get_permission(experiment_id, scorer_name, username)

    def list_scorer_permissions(self, username: str) -> List[ScorerPermission]:
        return self.scorer_repo.list_permissions_for_user(username)

    def update_scorer_permission(self, experiment_id: str, scorer_name: str, username: str, permission: str) -> ScorerPermission:
        return self.scorer_repo.update_permission(experiment_id, scorer_name, username, permission)

    def delete_scorer_permission(self, experiment_id: str, scorer_name: str, username: str) -> None:
        return self.scorer_repo.revoke_permission(experiment_id, scorer_name, username)

    def delete_scorer_permissions_for_scorer(self, experiment_id: str, scorer_name: str) -> None:
        """Delete all stored permissions for a scorer.

        This is used when a scorer is deleted. Unlike experiment permissions (keyed by UUID),
        scorer permissions are keyed by (experiment_id, scorer_name, subject). If the scorer
        is later recreated with the same name, stale permission rows could either conflict
        with inserts or unintentionally grant access.

        Args:
            experiment_id: The experiment ID owning the scorer.
            scorer_name: The scorer name.
        """

        from mlflow_oidc_auth.db.models import SqlScorerGroupPermission, SqlScorerPermission

        with self.ManagedSessionMaker() as session:
            session.query(SqlScorerPermission).filter(
                SqlScorerPermission.experiment_id == experiment_id,
                SqlScorerPermission.scorer_name == scorer_name,
            ).delete(synchronize_session=False)
            session.query(SqlScorerGroupPermission).filter(
                SqlScorerGroupPermission.experiment_id == experiment_id,
                SqlScorerGroupPermission.scorer_name == scorer_name,
            ).delete(synchronize_session=False)
            session.flush()

    # Scorer permissions (group-scoped)
    def create_group_scorer_permission(self, group_name: str, experiment_id: str, scorer_name: str, permission: str):
        return self.scorer_group_repo.grant_group_permission(group_name, experiment_id, scorer_name, permission)

    def update_group_scorer_permission(self, group_name: str, experiment_id: str, scorer_name: str, permission: str) -> ScorerGroupRegexPermission:
        return self.scorer_group_repo.update_group_permission(group_name, experiment_id, scorer_name, permission)

    def delete_group_scorer_permission(self, group_name: str, experiment_id: str, scorer_name: str) -> None:
        return self.scorer_group_repo.revoke_group_permission(group_name, experiment_id, scorer_name)

    def list_group_scorer_permissions(self, group_name: str):
        return self.scorer_group_repo.list_permissions_for_group(group_name)

    def get_user_groups_scorer_permission(self, experiment_id: str, scorer_name: str, username: str):
        return self.scorer_group_repo.get_group_permission_for_user_scorer(experiment_id, scorer_name, username)

    # Scorer regex (user-scoped)
    def create_scorer_regex_permission(self, regex: str, priority: int, permission: str, username: str) -> ScorerRegexPermission:
        return self.scorer_regex_repo.grant(regex=regex, priority=priority, permission=permission, username=username)

    def list_scorer_regex_permissions(self, username: str) -> List[ScorerRegexPermission]:
        return self.scorer_regex_repo.list_regex_for_user(username)

    def get_scorer_regex_permission(self, username: str, id: int) -> ScorerRegexPermission:
        return self.scorer_regex_repo.get(username=username, id=id)

    def update_scorer_regex_permission(self, id: int, regex: str, priority: int, permission: str, username: str) -> ScorerRegexPermission:
        return self.scorer_regex_repo.update(id=id, regex=regex, priority=priority, permission=permission, username=username)

    def delete_scorer_regex_permission(self, id: int, username: str) -> None:
        return self.scorer_regex_repo.revoke(id=id, username=username)

    # Scorer regex (group-scoped)
    def create_group_scorer_regex_permission(self, group_name: str, regex: str, priority: int, permission: str) -> ScorerGroupRegexPermission:
        return self.scorer_group_regex_repo.grant(group_name=group_name, regex=regex, priority=priority, permission=permission)

    def list_group_scorer_regex_permissions_for_groups_ids(self, group_ids: List[int]) -> List[ScorerGroupRegexPermission]:
        return self.scorer_group_regex_repo.list_permissions_for_groups_ids(group_ids)

    def list_group_scorer_regex_permissions(self, group_name: str) -> List[ScorerGroupRegexPermission]:
        from mlflow_oidc_auth.db.models import SqlGroup

        with self.ManagedSessionMaker() as session:
            group = session.query(SqlGroup).filter(SqlGroup.group_name == group_name).one_or_none()
            if group is None:
                return []
            return self.scorer_group_regex_repo.list_permissions_for_groups_ids([group.id])

    def get_group_scorer_regex_permission(self, group_name: str, id: int) -> ScorerGroupRegexPermission:
        return self.scorer_group_regex_repo.get(group_name=group_name, id=id)

    def update_group_scorer_regex_permission(self, id: int, group_name: str, regex: str, priority: int, permission: str) -> ScorerGroupRegexPermission:
        return self.scorer_group_regex_repo.update(id=id, group_name=group_name, regex=regex, priority=priority, permission=permission)

    def delete_group_scorer_regex_permission(self, id: int, group_name: str) -> None:
        return self.scorer_group_regex_repo.revoke(id=id, group_name=group_name)

    def authenticate_user(self, username: str, password: str) -> bool:
        return self.user_repo.authenticate(username, password)

    def create_user(self, username: str, password: str, display_name: str, is_admin: bool = False, is_service_account=False):
        return self.user_repo.create(username, password, display_name, is_admin, is_service_account)

    def has_user(self, username: str) -> bool:
        return self.user_repo.exist(username)

    def get_user(self, username: str) -> User:
        return self.user_repo.get(username)

    def get_user_profile(self, username: str) -> User:
        """Return a lightweight user entity for UI/admin checks.

        Unlike `get_user`, this method avoids loading permission collections.
        It is suitable for endpoints that only need basic user metadata.
        """

        return self.user_repo.get_profile(username)

    def list_users(self, is_service_account: bool = False, all: bool = False) -> List[User]:
        return self.user_repo.list(is_service_account, all)

    def update_user(
        self,
        username: str,
        password: Optional[str] = None,
        password_expiration: Optional[datetime] = None,
        is_admin: Optional[bool] = None,
        is_service_account: Optional[bool] = None,
    ) -> User:
        return self.user_repo.update(
            username=username,
            password=password,
            password_expiration=password_expiration,
            is_admin=is_admin,
            is_service_account=is_service_account,
        )

    def delete_user(self, username: str):
        return self.user_repo.delete(username)

    def create_experiment_permission(self, experiment_id: str, username: str, permission: str) -> ExperimentPermission:
        return self.experiment_repo.grant_permission(experiment_id, username, permission)

    def get_experiment_permission(self, experiment_id: str, username: str) -> ExperimentPermission:
        return self.experiment_repo.get_permission(experiment_id, username)

    def get_user_groups_experiment_permission(self, experiment_id: str, username: str) -> ExperimentPermission:
        return self.experiment_group_repo.get_group_permission_for_user_experiment(experiment_id, username)

    def list_experiment_permissions(self, username: str) -> List[ExperimentPermission]:
        return self.experiment_repo.list_permissions_for_user(username)

    def list_group_experiment_permissions(self, group_name: str) -> List[ExperimentPermission]:
        return self.experiment_group_repo.list_permissions_for_group(group_name)

    def list_group_id_experiment_permissions(self, group_id: int) -> List[ExperimentPermission]:
        return self.experiment_group_repo.list_permissions_for_group_id(group_id)

    def list_user_groups_experiment_permissions(self, username: str) -> List[ExperimentPermission]:
        return self.experiment_group_repo.list_permissions_for_user_groups(username)

    def update_experiment_permission(self, experiment_id: str, username: str, permission: str) -> ExperimentPermission:
        return self.experiment_repo.update_permission(experiment_id, username, permission)

    def delete_experiment_permission(self, experiment_id: str, username: str):
        return self.experiment_repo.revoke_permission(experiment_id, username)

    def create_registered_model_permission(self, name: str, username: str, permission: str) -> RegisteredModelPermission:
        return self.registered_model_repo.create(name, username, permission)

    def get_registered_model_permission(self, name: str, username: str) -> RegisteredModelPermission:
        return self.registered_model_repo.get(name, username)

    def get_user_groups_registered_model_permission(self, name: str, username: str) -> RegisteredModelPermission:
        return self.registered_model_group_repo.get_for_user(name, username)

    def list_registered_model_permissions(self, username: str) -> List[RegisteredModelPermission]:
        return self.registered_model_repo.list_for_user(username)

    def list_user_groups_registered_model_permissions(self, username: str) -> List[RegisteredModelPermission]:
        return self.registered_model_group_repo.list_for_user(username)

    def update_registered_model_permission(self, name: str, username: str, permission: str) -> RegisteredModelPermission:
        return self.registered_model_repo.update(name, username, permission)

    def rename_registered_model_permissions(self, old_name: str, new_name: str):
        return self.registered_model_repo.rename(old_name, new_name)

    def delete_registered_model_permission(self, name: str, username: str):
        return self.registered_model_repo.delete(name, username)

    def wipe_registered_model_permissions(self, name: str):
        return self.registered_model_repo.wipe(name)

    def list_experiment_permissions_for_experiment(self, experiment_id: str) -> List[ExperimentPermission]:
        return self.experiment_repo.list_permissions_for_experiment(experiment_id)

    def populate_groups(self, group_names: List[str]):
        return self.group_repo.create_groups(group_names)

    def get_groups(self) -> List[str]:
        return self.group_repo.list_groups()

    def get_group_users(self, group_name: str) -> List[User]:
        return self.group_repo.list_group_members(group_name)

    def add_user_to_group(self, username: str, group_name: str) -> None:
        return self.group_repo.add_user_to_group(username, group_name)

    def remove_user_from_group(self, username: str, group_name: str) -> None:
        return self.group_repo.remove_user_from_group(username, group_name)

    def get_groups_for_user(self, username: str) -> List[str]:
        return self.group_repo.list_groups_for_user(username)

    def get_groups_ids_for_user(self, username: str) -> List[int]:
        return self.group_repo.list_group_ids_for_user(username)

    def set_user_groups(self, username: str, group_names: List[str]) -> None:
        return self.group_repo.set_groups_for_user(username, group_names)

    def get_group_experiments(self, group_name: str) -> List[ExperimentPermission]:
        return self.experiment_group_repo.list_permissions_for_group(group_name)

    def create_group_experiment_permission(self, group_name: str, experiment_id: str, permission: str) -> ExperimentPermission:
        return self.experiment_group_repo.grant_group_permission(group_name, experiment_id, permission)

    def delete_group_experiment_permission(self, group_name: str, experiment_id: str) -> None:
        return self.experiment_group_repo.revoke_group_permission(group_name, experiment_id)

    def update_group_experiment_permission(self, group_name: str, experiment_id: str, permission: str) -> ExperimentPermission:
        return self.experiment_group_repo.update_group_permission(group_name, experiment_id, permission)

    def get_group_models(self, group_name: str) -> List[RegisteredModelPermission]:
        return self.registered_model_group_repo.get(group_name)

    def create_group_model_permission(self, group_name: str, name: str, permission: str):
        return self.registered_model_group_repo.create(group_name, name, permission)

    def rename_group_model_permissions(self, old_name: str, new_name: str):
        return self.registered_model_group_repo.rename(old_name, new_name)

    def delete_group_model_permission(self, group_name: str, name: str):
        return self.registered_model_group_repo.delete(group_name, name)

    def wipe_group_model_permissions(self, name: str):
        return self.registered_model_group_repo.wipe(name)

    def update_group_model_permission(self, group_name: str, name: str, permission: str):
        return self.registered_model_group_repo.update(group_name, name, permission)

    # Prompt CRUD
    def create_group_prompt_permission(self, group_name: str, name: str, permission: str):
        return self.prompt_group_repo.grant_prompt_permission_to_group(group_name, name, permission)

    def get_group_prompts(self, group_name: str) -> List[RegisteredModelPermission]:
        return self.prompt_group_repo.list_prompt_permissions_for_group(group_name)

    def update_group_prompt_permission(self, group_name: str, name: str, permission: str):
        return self.prompt_group_repo.update_prompt_permission_for_group(group_name, name, permission)

    def delete_group_prompt_permission(self, group_name: str, name: str):
        return self.prompt_group_repo.revoke_prompt_permission_from_group(group_name, name)

    # Experiment regex CRUD
    def create_experiment_regex_permission(self, regex: str, priority: int, permission: str, username: str):
        return self.experiment_regex_repo.grant(regex, priority, permission, username)

    def get_experiment_regex_permission(self, username: str, id: int) -> ExperimentRegexPermission:
        return self.experiment_regex_repo.get(username=username, id=id)

    def list_experiment_regex_permissions(self, username: str) -> List[ExperimentRegexPermission]:
        return self.experiment_regex_repo.list_regex_for_user(username)

    def update_experiment_regex_permission(self, regex: str, priority: int, permission: str, username: str, id: int) -> ExperimentRegexPermission:
        return self.experiment_regex_repo.update(regex=regex, priority=priority, permission=permission, username=username, id=id)

    def delete_experiment_regex_permission(self, username: str, id: int) -> None:
        return self.experiment_regex_repo.revoke(username=username, id=id)

    # Experiment regex group CRUD
    def create_group_experiment_regex_permission(self, group_name: str, regex: str, priority: int, permission: str) -> ExperimentGroupRegexPermission:
        return self.experiment_group_regex_repo.grant(group_name, regex, priority, permission)

    def get_group_experiment_regex_permission(self, group_name: str, id: int) -> ExperimentGroupRegexPermission:
        return self.experiment_group_regex_repo.get(group_name, id)

    def list_group_experiment_regex_permissions(self, group_name: str) -> List[ExperimentGroupRegexPermission]:
        return self.experiment_group_regex_repo.list_permissions_for_group(group_name)

    def list_group_experiment_regex_permissions_for_groups(self, group_names: List[str]) -> List[ExperimentGroupRegexPermission]:
        return self.experiment_group_regex_repo.list_permissions_for_groups(group_names)

    def list_group_experiment_regex_permissions_for_groups_ids(self, group_ids: List[int]) -> List[ExperimentGroupRegexPermission]:
        return self.experiment_group_regex_repo.list_permissions_for_groups_ids(group_ids)

    def update_group_experiment_regex_permission(self, id: int, group_name: str, regex: str, priority: int, permission: str) -> ExperimentGroupRegexPermission:
        return self.experiment_group_regex_repo.update(id, group_name, regex, priority, permission)

    def delete_group_experiment_regex_permission(self, group_name: str, id: int) -> None:
        return self.experiment_group_regex_repo.revoke(group_name, id)

    # Registered model regex CRUD
    def create_registered_model_regex_permission(self, regex: str, priority: int, permission: str, username: str):
        return self.registered_model_regex_repo.grant(regex, priority, permission, username)

    def get_registered_model_regex_permission(self, id: int, username: str) -> RegisteredModelRegexPermission:
        return self.registered_model_regex_repo.get(id, username)

    def list_registered_model_regex_permissions(self, username: str) -> List[RegisteredModelRegexPermission]:
        return self.registered_model_regex_repo.list_regex_for_user(username)

    def update_registered_model_regex_permission(self, id: int, regex: str, priority: int, permission: str, username: str) -> RegisteredModelRegexPermission:
        return self.registered_model_regex_repo.update(id, regex, priority, permission, username)

    def delete_registered_model_regex_permission(self, id: int, username: str) -> None:
        return self.registered_model_regex_repo.revoke(id, username)

    # Registered model regex group CRUD
    def create_group_registered_model_regex_permission(
        self, group_name: str, regex: str, priority: int, permission: str
    ) -> RegisteredModelGroupRegexPermission:
        return self.registered_model_group_regex_repo.grant(group_name=group_name, regex=regex, priority=priority, permission=permission)

    def get_group_registered_model_regex_permission(self, group_name: str, id: int) -> RegisteredModelGroupRegexPermission:
        return self.registered_model_group_regex_repo.get(id=id, group_name=group_name)

    def list_group_registered_model_regex_permissions(self, group_name: str) -> List[RegisteredModelGroupRegexPermission]:
        return self.registered_model_group_regex_repo.list_permissions_for_group(group_name)

    def list_group_registered_model_regex_permissions_for_groups(self, group_names: List[str]) -> List[RegisteredModelGroupRegexPermission]:
        return self.registered_model_group_regex_repo.list_permissions_for_groups(group_names)

    def list_group_registered_model_regex_permissions_for_groups_ids(self, group_ids: List[int]) -> List[RegisteredModelGroupRegexPermission]:
        return self.registered_model_group_regex_repo.list_permissions_for_groups_ids(group_ids)

    def update_group_registered_model_regex_permission(
        self, id: int, group_name: str, regex: str, priority: int, permission: str
    ) -> RegisteredModelGroupRegexPermission:
        return self.registered_model_group_regex_repo.update(id=id, group_name=group_name, regex=regex, priority=priority, permission=permission)

    def delete_group_registered_model_regex_permission(self, group_name: str, id: int) -> None:
        return self.registered_model_group_regex_repo.revoke(group_name=group_name, id=id)

    # Prompt regex CRUD
    def create_prompt_regex_permission(self, regex: str, priority: int, permission: str, username: str, prompt: bool = True):
        return self.prompt_regex_repo.grant(regex=regex, priority=priority, permission=permission, username=username, prompt=prompt)

    def get_prompt_regex_permission(self, id: int, username: str, prompt: bool = True) -> RegisteredModelRegexPermission:
        return self.prompt_regex_repo.get(id=id, username=username, prompt=prompt)

    def list_prompt_regex_permissions(self, username: str, prompt: bool = True) -> List[RegisteredModelRegexPermission]:
        return self.prompt_regex_repo.list_regex_for_user(username=username, prompt=prompt)

    def update_prompt_regex_permission(
        self, id: int, regex: str, priority: int, permission: str, username: str, prompt: bool = True
    ) -> RegisteredModelRegexPermission:
        return self.prompt_regex_repo.update(id=id, regex=regex, priority=priority, permission=permission, username=username, prompt=prompt)

    def delete_prompt_regex_permission(self, id: int, username: str) -> None:
        return self.prompt_regex_repo.revoke(id=id, username=username, prompt=True)

    # Prompt regex group CRUD
    def create_group_prompt_regex_permission(self, regex: str, priority: int, permission: str, group_name: str, prompt: bool = True):
        return self.prompt_group_regex_repo.grant(regex=regex, priority=priority, permission=permission, group_name=group_name, prompt=prompt)

    def get_group_prompt_regex_permission(self, id: int, group_name: str, prompt: bool = True) -> RegisteredModelGroupRegexPermission:
        return self.prompt_group_regex_repo.get(id=id, group_name=group_name, prompt=prompt)

    def list_group_prompt_regex_permissions(self, group_name: str, prompt: bool = True) -> List[RegisteredModelGroupRegexPermission]:
        return self.prompt_group_regex_repo.list_permissions_for_group(group_name=group_name, prompt=prompt)

    def list_group_prompt_regex_permissions_for_groups(self, group_names: List[str], prompt: bool = True) -> List[RegisteredModelGroupRegexPermission]:
        return self.prompt_group_regex_repo.list_permissions_for_groups(group_names=group_names, prompt=prompt)

    def list_group_prompt_regex_permissions_for_groups_ids(self, group_ids: List[int], prompt: bool = True) -> List[RegisteredModelGroupRegexPermission]:
        return self.prompt_group_regex_repo.list_permissions_for_groups_ids(group_ids=group_ids, prompt=prompt)

    def update_group_prompt_regex_permission(
        self, id: int, regex: str, priority: int, permission: str, group_name: str, prompt: bool = True
    ) -> RegisteredModelGroupRegexPermission:
        return self.prompt_group_regex_repo.update(id=id, regex=regex, priority=priority, permission=permission, group_name=group_name, prompt=prompt)

    def delete_group_prompt_regex_permission(self, id: int, group_name: str) -> None:
        return self.prompt_group_regex_repo.revoke(id=id, group_name=group_name, prompt=True)



from mlflow_oidc_auth.hooks.after_request import after_request_hook
from mlflow_oidc_auth.hooks.before_request import before_request_hook


__all__ = [
    "before_request_hook",
    "after_request_hook",
]



"""
URI construction and validation utilities for MLflow OIDC Auth.

This module provides functionality to dynamically construct OIDC redirect URIs
and other URI-related operations based on the current request context. With ProxyFix
middleware configured, FastAPI's request object automatically contains the correct
values from proxy headers.

Key Features:
- Dynamic OIDC redirect URI construction
- URL normalization (port handling)
- Proxy-aware URI building
- Request context utilities

Dependencies:
- FastAPI request context (requires active request)
- ProxyFix middleware for proper proxy header handling
"""

from typing import Optional
from urllib.parse import urlparse, urlunparse


def normalize_url_port(url: str) -> str:
    """
    Normalize URL by removing standard ports (80 for HTTP, 443 for HTTPS).

    This ensures that OIDC redirect URIs don't include redundant port numbers
    that might cause mismatches with provider configurations. Standard ports
    are automatically omitted while custom ports are preserved.

    Args:
        url (str): The URL to normalize

    Returns:
        str: URL with standard ports removed

    Examples:
        >>> normalize_url_port("http://example.com:80/path")
        "http://example.com/path"
        >>> normalize_url_port("https://example.com:443/path")
        "https://example.com/path"
        >>> normalize_url_port("https://example.com:8443/path")
        "https://example.com:8443/path"

    Edge Cases:
        - Invalid URLs: Returns original URL unchanged
        - URLs without ports: Returns original URL unchanged
        - Non-standard ports: Preserves the port in the URL
        - None input: Raises TypeError
        - Empty string: Returns empty string unchanged
    """
    # Handle None input explicitly
    if url is None:
        raise TypeError("normalize_url_port() missing 1 required positional argument: 'url'")

    # Handle empty string
    if not url:
        return url

    try:
        parsed = urlparse(url)

        # Check if port should be omitted (standard ports only)
        should_omit_port = (parsed.scheme == "http" and parsed.port == 80) or (parsed.scheme == "https" and parsed.port == 443)

        if should_omit_port:
            # Reconstruct URL without port by setting netloc without port
            # netloc format: [userinfo@]host[:port]
            if "@" in parsed.netloc:
                userinfo, host_port = parsed.netloc.rsplit("@", 1)
                host = host_port.split(":")[0]
                new_netloc = f"{userinfo}@{host}"
            else:
                host = parsed.netloc.split(":")[0]
                new_netloc = host

            # Create new parsed URL with modified netloc
            normalized = parsed._replace(netloc=new_netloc)
            return urlunparse(normalized)

        return url

    except (ValueError, AttributeError) as e:
        # Handle malformed URLs gracefully by returning original
        from flask import current_app

        if current_app:
            current_app.logger.warning(f"Failed to normalize URL '{url}': {e}")
        return url


from fastapi import Request


def _get_base_url_from_request(request: Request) -> str:
    """
    Extract the base URL from the current FastAPI request context.

    With ProxyFix middleware configured, this function automatically handles
    proxy headers (X-Forwarded-Proto, X-Forwarded-Host, X-Forwarded-Prefix)
    to construct the correct base URL regardless of proxy configuration.

    Returns:
        str: The normalized base URL for the current request

    Examples:
        Direct access: "http://localhost:5000"
        Behind proxy: "https://example.com/my-app"

    Edge Cases:
        - No active request context: Raises RuntimeError
        - Invalid request data: Returns best-effort URL construction

    Note:
        This function requires an active FastAPI request context and should
        only be called during request processing.
    """
    if request is None:
        raise RuntimeError("_get_base_url_from_request() requires an active FastAPI request context")

    parsed_url = urlparse(str(request.url))
    # Use root_path for the base path (proxy prefix), default to "" if empty
    base_path = request.scope.get("root_path", "")

    # Reconstruct the base URL with the correct base path
    base_url_parts = (parsed_url.scheme, parsed_url.netloc, base_path, "", "", "")
    raw_base_url = urlunparse(base_url_parts)

    # Normalize to remove standard ports for consistency
    return normalize_url_port(raw_base_url)


def _get_dynamic_redirect_uri(request: Request, callback_path: str) -> str:
    """
    Dynamically construct the OIDC redirect URI based on the current request context.

    With ProxyFix middleware configured, Flask's request object automatically
    contains the correct scheme, host, and URL from X-Forwarded-* headers.
    This allows the redirect URI to adapt automatically to different proxy
    configurations without requiring manual configuration.

    Args:
        callback_path (str): The callback path to append to the base URL.
                           Defaults to the value from mlflow_oidc_auth.routes.CALLBACK

    Returns:
        str: The complete OIDC redirect URI

    Examples:
        Direct access: "http://localhost:5000/callback"
        Behind proxy: "https://example.com/my-app/callback"
        Custom callback: "https://example.com/my-app/auth/callback"

    Edge Cases:
        - callback_path without leading slash: Automatically adds leading slash
        - Empty callback_path: Uses "/" as callback path
        - No active request context: Raises RuntimeError

    Note:
        This function requires an active Flask request context and should
        only be called during OIDC authentication flow.
    """
    base_url = _get_base_url_from_request(request=request)

    # Ensure callback path starts with /
    if not callback_path:
        callback_path = "/"
    elif not callback_path.startswith("/"):
        callback_path = f"/{callback_path}"

    redirect_uri = f"{base_url}{callback_path}"

    return redirect_uri


def get_configured_or_dynamic_redirect_uri(request: Request, callback_path: str, configured_uri: Optional[str]) -> str:
    """
    Get the OIDC redirect URI, using configured value if available, otherwise calculate dynamically.

    This function provides a fallback mechanism that uses a manually configured
    redirect URI if available, but falls back to dynamic calculation if not set.
    This approach provides flexibility for both automated deployments and cases
    where manual override is needed.

    Args:
        configured_uri (Optional[str]): The manually configured redirect URI (can be None)
        callback_path (str): The callback path for dynamic calculation.
                           Defaults to the value from mlflow_oidc_auth.routes.CALLBACK

    Returns:
        str: The OIDC redirect URI to use
    """
    # Use configured URI if it's provided and not empty/whitespace
    if configured_uri and configured_uri.strip():
        return configured_uri.strip()

    # Fall back to dynamic calculation
    return _get_dynamic_redirect_uri(request=request, callback_path=callback_path)



from unittest.mock import MagicMock, patch

import pytest

from mlflow_oidc_auth.validators import run


class DummyPermission:
    def __init__(self, can_read=False, can_update=False, can_delete=False, can_manage=False):
        self.can_read = can_read
        self.can_update = can_update
        self.can_delete = can_delete
        self.can_manage = can_manage


def _patch_permission(**kwargs):
    return patch(
        "mlflow_oidc_auth.validators.run.effective_experiment_permission",
        return_value=MagicMock(permission=DummyPermission(**kwargs)),
    )


def test__get_permission_from_run_id():
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store, patch(
        "mlflow_oidc_auth.validators.run.effective_experiment_permission",
        return_value=MagicMock(permission=DummyPermission(can_read=True)),
    ):
        mock_store.return_value.get_run.return_value = mock_run
        perm = run._get_permission_from_run_id("alice")
        assert perm.can_read is True


def test_validate_can_read_run():
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run("alice") is True


def test_validate_can_update_run():
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_update=True):
            assert run.validate_can_update_run("alice") is True


def test_validate_can_delete_run():
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_delete=True):
            assert run.validate_can_delete_run("alice") is True


def test_validate_can_manage_run():
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_manage=True):
            assert run.validate_can_manage_run("alice") is True


# Additional tests for missing coverage and edge cases


def test__get_permission_from_run_id_no_permission():
    """Test when user has no permissions for run"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store, patch(
        "mlflow_oidc_auth.validators.run.effective_experiment_permission",
        return_value=MagicMock(permission=DummyPermission()),
    ):
        mock_store.return_value.get_run.return_value = mock_run
        perm = run._get_permission_from_run_id("alice")
        assert perm.can_read is False
        assert perm.can_update is False
        assert perm.can_delete is False
        assert perm.can_manage is False


def test_validate_can_read_run_false():
    """Test when user cannot read run"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=False):
            assert run.validate_can_read_run("alice") is False


def test_validate_can_update_run_false():
    """Test when user cannot update run"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_update=False):
            assert run.validate_can_update_run("alice") is False


def test_validate_can_delete_run_false():
    """Test when user cannot delete run"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_delete=False):
            assert run.validate_can_delete_run("alice") is False


def test_validate_can_manage_run_false():
    """Test when user cannot manage run"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_manage=False):
            assert run.validate_can_manage_run("alice") is False


# Security and edge case tests


def test_validate_with_none_username_run():
    """Test validation functions with None username"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run(None) is True


def test_validate_with_empty_username_run():
    """Test validation functions with empty username"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run("") is True


def test_validate_with_special_characters_username_run():
    """Test validation functions with special characters in username"""
    username = "user@domain.com"
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run(username) is True


def test_validate_with_malformed_run_id():
    """Test with malformed run ID"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value=""), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run("alice") is True


def test_validate_with_very_long_run_id():
    """Test with very long run ID"""
    long_run_id = "run_" + "a" * 1000
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value=long_run_id), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run("alice") is True


def test_get_run_store_exception():
    """Test when store raises an exception for run"""
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.side_effect = Exception("Store error")

        with pytest.raises(Exception, match="Store error"):
            run._get_permission_from_run_id("alice")


def test_permission_inheritance_scenarios_run():
    """Test various permission inheritance scenarios for runs"""
    mock_run = MagicMock()
    mock_run.info.experiment_id = "exp1"
    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run
        # Test partial permissions
        with _patch_permission(can_read=True, can_update=False, can_delete=False, can_manage=False):
            assert run.validate_can_read_run("alice") is True
            assert run.validate_can_update_run("alice") is False
            assert run.validate_can_delete_run("alice") is False
            assert run.validate_can_manage_run("alice") is False


def test_run_with_different_experiment_ids():
    """Test runs with different experiment IDs"""
    # Test with numeric experiment ID
    mock_run1 = MagicMock()
    mock_run1.info.experiment_id = "123"

    # Test with string experiment ID
    mock_run2 = MagicMock()
    mock_run2.info.experiment_id = "default"

    with patch("mlflow_oidc_auth.validators.run.get_request_param", return_value="run123"), patch(
        "mlflow_oidc_auth.validators.run._get_tracking_store"
    ) as mock_store:
        mock_store.return_value.get_run.return_value = mock_run1
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run("alice") is True

        mock_store.return_value.get_run.return_value = mock_run2
        with _patch_permission(can_read=True):
            assert run.validate_can_read_run("alice") is True



from typing import Optional

from pydantic import BaseModel


class CreateAccessTokenRequest(BaseModel):
    """Request model for creating access tokens."""

    username: Optional[str] = None  # Optional, will use authenticated user if not provided
    expiration: Optional[str] = None  # ISO 8601 format string


class CreateUserRequest(BaseModel):
    """Request model for creating users."""

    username: str
    display_name: str
    is_admin: bool = False
    is_service_account: bool = False



"""
Application configuration for MLflow OIDC Auth Plugin.

This module provides configuration management with support for multiple
configuration sources (environment variables, AWS Secrets Manager, Azure Key Vault,
Kubernetes Secrets, etc.) through the pluggable provider system.

Environment Variables for Provider Configuration:
    CONFIG_AWS_SECRETS_ENABLED: Enable AWS Secrets Manager provider
    CONFIG_AWS_PARAMETER_STORE_ENABLED: Enable AWS Parameter Store provider
    CONFIG_AZURE_KEYVAULT_ENABLED: Enable Azure Key Vault provider
    CONFIG_VAULT_ENABLED: Enable HashiCorp Vault provider
    CONFIG_K8S_SECRETS_ENABLED: Enable Kubernetes Secrets provider
    CONFIG_PROVIDERS: Comma-separated list of providers to use (optional filter)

See config_providers/ for detailed configuration of each provider.
"""

import secrets

from dotenv import load_dotenv

from mlflow_oidc_auth.config_providers import config_manager
from mlflow_oidc_auth.logger import get_logger

load_dotenv()  # take environment variables from .env.
logger = get_logger()


def get_bool_env_variable(variable: str, default_value: bool) -> bool:
    """Get a boolean value from configuration.

    Parameters:
        variable: The configuration key name.
        default_value: Default value if not found.

    Returns:
        Boolean value parsed from configuration.
    """
    return config_manager.get_bool(variable, default=default_value)


class AppConfig:
    """Application configuration container.

    This class loads configuration from the pluggable provider system,
    which supports multiple sources in priority order:
        1. Cloud providers (AWS Secrets Manager, Azure Key Vault, etc.)
        2. Kubernetes Secrets (mounted as files)
        3. Environment variables (fallback)

    Attributes:
        DEFAULT_MLFLOW_PERMISSION: Default permission level for new resources.
        SECRET_KEY: Secret key for session management.
        OIDC_USERS_DB_URI: Database URI for user/permission storage.
        OIDC_CLIENT_ID: OAuth client ID.
        OIDC_CLIENT_SECRET: OAuth client secret (sensitive).
        ... and more (see source for full list)
    """

    def __init__(self) -> None:
        """Initialize configuration from the provider chain."""
        # Permission settings
        self.DEFAULT_MLFLOW_PERMISSION = config_manager.get("DEFAULT_MLFLOW_PERMISSION", "MANAGE")
        self.PERMISSION_SOURCE_ORDER = config_manager.get_list("PERMISSION_SOURCE_ORDER", default=["user", "group", "regex", "group-regex"])

        # Security settings (secrets - may come from Secrets Manager/Key Vault)
        self.SECRET_KEY = config_manager.get("SECRET_KEY") or secrets.token_hex(16)
        self.OIDC_CLIENT_SECRET = config_manager.get("OIDC_CLIENT_SECRET")

        # Database settings (sensitive)
        self.OIDC_USERS_DB_URI = config_manager.get("OIDC_USERS_DB_URI", "sqlite:///auth.db")

        # OIDC provider settings
        self.OIDC_DISCOVERY_URL = config_manager.get("OIDC_DISCOVERY_URL")
        self.OIDC_CLIENT_ID = config_manager.get("OIDC_CLIENT_ID")
        # OIDC_REDIRECT_URI: If not set, will be calculated dynamically based on request headers
        # This enables automatic proxy path detection for OIDC callbacks
        self.OIDC_REDIRECT_URI = config_manager.get("OIDC_REDIRECT_URI")
        self.OIDC_SCOPE = config_manager.get("OIDC_SCOPE", "openid,email,profile")
        self.OIDC_PROVIDER_DISPLAY_NAME = config_manager.get("OIDC_PROVIDER_DISPLAY_NAME", "Login with OIDC")
        self.OIDC_GROUPS_ATTRIBUTE = config_manager.get("OIDC_GROUPS_ATTRIBUTE", "groups")

        # Group settings
        self.OIDC_GROUP_NAME = config_manager.get_list("OIDC_GROUP_NAME", default=["mlflow"])
        self.OIDC_ADMIN_GROUP_NAME = config_manager.get_list("OIDC_ADMIN_GROUP_NAME", default=["mlflow-admin"])
        self.OIDC_GROUP_DETECTION_PLUGIN = config_manager.get("OIDC_GROUP_DETECTION_PLUGIN")

        # Database migration settings
        self.OIDC_ALEMBIC_VERSION_TABLE = config_manager.get("OIDC_ALEMBIC_VERSION_TABLE", "alembic_version")

        # UI settings
        self.EXTEND_MLFLOW_MENU = config_manager.get_bool("EXTEND_MLFLOW_MENU", default=True)
        self.DEFAULT_LANDING_PAGE_IS_PERMISSIONS = config_manager.get_bool("DEFAULT_LANDING_PAGE_IS_PERMISSIONS", default=True)
        self.AUTOMATIC_LOGIN_REDIRECT = config_manager.get_bool("AUTOMATIC_LOGIN_REDIRECT", default=False)

    def refresh(self) -> None:
        """Reload configuration from all providers.

        Call this method to reload configuration after secret rotation
        or configuration changes.
        """
        config_manager.refresh()
        self.__init__()


config = AppConfig()



"""Admin capabilities integration tests for MLflow OIDC Auth Plugin.

Test IDs: ADM-001 through ADM-022

Tests verify:
- Admin can manage permissions on any resource
- Admin can manage users and service accounts
- Admin bypasses all permission restrictions
"""

from __future__ import annotations

import time
import uuid
from urllib.parse import quote

import httpx
import pytest


# =============================================================================
# Helper Functions
# =============================================================================


def _create_experiment(client: httpx.Client, experiment_name: str) -> tuple[bool, str]:
    """Create an experiment and return (success, experiment_id)."""
    get_api = f"ajax-api/2.0/mlflow/experiments/get-by-name?experiment_name={quote(experiment_name)}"
    create_api = "ajax-api/2.0/mlflow/experiments/create"

    get_resp = client.get(get_api)
    if get_resp.status_code == 200:
        payload = get_resp.json()
        exp = payload.get("experiment", payload)
        exp_id = exp.get("experiment_id") or exp.get("experimentId") or exp.get("id")
        return True, str(exp_id)

    create_resp = client.post(create_api, json={"name": experiment_name})
    if create_resp.status_code != 200:
        return False, f"Create failed: {create_resp.status_code}"

    get_resp2 = client.get(get_api)
    if get_resp2.status_code == 200:
        payload2 = get_resp2.json()
        exp2 = payload2.get("experiment", payload2)
        exp_id2 = exp2.get("experiment_id") or exp2.get("experimentId") or exp2.get("id")
        return True, str(exp_id2)

    return False, "Failed to get experiment after creation"


def _create_model(client: httpx.Client, model_name: str) -> bool:
    """Create a registered model."""
    create_api = "ajax-api/2.0/mlflow/registered-models/create"
    create_resp = client.post(create_api, json={"name": model_name})
    return create_resp.status_code in (200, 409)


def _grant_user_permission(
    client: httpx.Client,
    resource_type: str,
    resource_id: str,
    target_user: str,
    permission: str,
) -> bool:
    """Grant user-level permission."""
    if resource_type == "experiment":
        api = f"api/2.0/mlflow/permissions/users/{quote(target_user)}/experiments/{quote(resource_id)}"
    elif resource_type == "model":
        api = f"api/2.0/mlflow/permissions/users/{quote(target_user)}/registered-models/{quote(resource_id)}"
    else:
        return False

    resp = client.patch(api, json={"permission": permission})
    if resp.status_code == 404:
        resp = client.post(api, json={"permission": permission})

    return resp.status_code in (200, 201)


def _is_ok(status_code: int) -> bool:
    """Check if response indicates success."""
    return status_code == 200


# =============================================================================
# Test Fixtures
# =============================================================================


@pytest.fixture(scope="module")
def admin_test_resources(
    base_url: str,
    ensure_server: None,
    http_client_factory,
    test_run_id: str,
):
    """Create resources owned by a regular user for admin tests.

    Alice creates resources and grants herself MANAGE,
    then grants specific limited permissions to Frank (admin) to verify bypass.
    """
    owner = "alice@example.com"
    admin = "frank@example.com"

    experiment_name = f"admin-test-exp-{test_run_id}"
    model_name = f"admin-test-model-{test_run_id}"

    # Resources with NO_PERMISSIONS for admin (to test bypass)
    restricted_exp_name = f"admin-restricted-exp-{test_run_id}"
    restricted_model_name = f"admin-restricted-model-{test_run_id}"

    with http_client_factory(owner) as client:
        # Create regular resources
        exp_success, experiment_id = _create_experiment(client, experiment_name)
        assert exp_success, f"Failed to create experiment: {experiment_id}"

        assert _create_model(client, model_name), "Failed to create model"

        # Create restricted resources and grant admin NO_PERMISSIONS
        rest_exp_success, restricted_exp_id = _create_experiment(client, restricted_exp_name)
        assert rest_exp_success, f"Failed to create restricted experiment"

        assert _create_model(client, restricted_model_name), "Failed to create restricted model"

        # Grant admin NO_PERMISSIONS on restricted resources (admin should bypass)
        _grant_user_permission(client, "experiment", restricted_exp_id, admin, "NO_PERMISSIONS")
        _grant_user_permission(client, "model", restricted_model_name, admin, "NO_PERMISSIONS")

        # Grant admin READ on regular resources (admin should still have full access)
        _grant_user_permission(client, "experiment", experiment_id, admin, "READ")
        _grant_user_permission(client, "model", model_name, admin, "READ")

    return {
        "owner": owner,
        "admin": admin,
        "experiment_name": experiment_name,
        "experiment_id": experiment_id,
        "model_name": model_name,
        "restricted_exp_name": restricted_exp_name,
        "restricted_exp_id": restricted_exp_id,
        "restricted_model_name": restricted_model_name,
    }


# =============================================================================
# ADM-001 to ADM-006: Admin Permission Management
# =============================================================================


@pytest.mark.integration
def test_admin_views_any_experiment(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-001: Admin can view any experiment."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        api = f"ajax-api/2.0/mlflow/experiments/get-by-name?experiment_name={quote(resources['experiment_name'])}"
        resp = client.get(api)
        assert _is_ok(resp.status_code), f"Admin should view any experiment: {resp.status_code}"


@pytest.mark.integration
def test_admin_modifies_any_experiment(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-002: Admin can modify any experiment (create runs)."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        api = "api/2.0/mlflow/runs/create"
        resp = client.post(
            api,
            json={
                "experiment_id": resources["experiment_id"],
                "start_time": int(time.time() * 1000),
                "tags": [{"key": "mlflow.runName", "value": f"admin-run-{uuid.uuid4().hex[:6]}"}],
            },
        )
        assert _is_ok(resp.status_code), f"Admin should create run: {resp.status_code}"


@pytest.mark.integration
def test_admin_deletes_any_experiment(
    http_client_factory,
    admin_test_resources,
    test_run_id: str,
) -> None:
    """ADM-003: Admin can delete any experiment."""
    admin = admin_test_resources["admin"]
    owner = admin_test_resources["owner"]

    # Create a new experiment to delete
    delete_exp_name = f"admin-delete-test-{test_run_id}-{uuid.uuid4().hex[:6]}"

    with http_client_factory(owner) as client:
        exp_success, exp_id = _create_experiment(client, delete_exp_name)
        assert exp_success, "Failed to create experiment for delete test"

    with http_client_factory(admin) as client:
        api = "ajax-api/2.0/mlflow/experiments/delete"
        resp = client.post(api, json={"experiment_id": exp_id})
        assert _is_ok(resp.status_code), f"Admin should delete experiment: {resp.status_code}"


@pytest.mark.integration
def test_admin_grants_permission_on_any_resource(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-004: Admin can grant permission on any resource."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        api = f"api/2.0/mlflow/permissions/users/{quote('newuser@example.com')}/experiments/{quote(resources['experiment_id'])}"
        resp = client.post(api, json={"permission": "READ"})

        if resp.status_code == 409:  # Already exists
            resp = client.patch(api, json={"permission": "READ"})

        assert resp.status_code in (200, 201), f"Admin should grant permission: {resp.status_code}"


@pytest.mark.integration
def test_admin_revokes_permission_on_any_resource(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-005: Admin can revoke permission on any resource."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        # First grant, then revoke
        grant_api = f"api/2.0/mlflow/permissions/users/{quote('newuser@example.com')}/experiments/{quote(resources['experiment_id'])}"

        # Grant first
        client.post(grant_api, json={"permission": "READ"})

        # Now revoke (delete)
        resp = client.request("DELETE", grant_api)
        assert resp.status_code in (200, 204, 404), f"Admin should revoke permission: {resp.status_code}"


@pytest.mark.integration
def test_admin_overrides_existing_permission(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-006: Admin can override existing permission."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        api = f"api/2.0/mlflow/permissions/users/{quote('bob@example.com')}/experiments/{quote(resources['experiment_id'])}"

        # Set initial permission
        client.post(api, json={"permission": "READ"})

        # Override to EDIT
        resp = client.patch(api, json={"permission": "EDIT"})
        assert resp.status_code in (200, 201), f"Admin should override permission: {resp.status_code}"


# =============================================================================
# ADM-010 to ADM-014: Admin User Management
# =============================================================================


@pytest.mark.integration
def test_admin_lists_all_users(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-010: Admin can list all users."""
    admin = admin_test_resources["admin"]

    with http_client_factory(admin) as client:
        api = "api/2.0/mlflow/users"
        resp = client.get(api)
        assert _is_ok(resp.status_code), f"Admin should list users: {resp.status_code}"

        data = resp.json()
        assert "users" in data or isinstance(data, list), "Response should contain users"


@pytest.mark.integration
def test_admin_creates_service_account(
    http_client_factory,
    admin_test_resources,
    test_run_id: str,
) -> None:
    """ADM-011: Admin can create service account."""
    admin = admin_test_resources["admin"]
    svc_username = f"svc-admin-test-{test_run_id}@example.com"

    with http_client_factory(admin) as client:
        api = "api/2.0/mlflow/users"
        resp = client.post(
            api,
            json={
                "username": svc_username,
                "display_name": f"Admin Test Service Account {test_run_id}",
                "is_admin": False,
                "is_service_account": True,
            },
        )
        assert resp.status_code in (200, 201), f"Admin should create service account: {resp.status_code}"


@pytest.mark.integration
def test_admin_creates_token_for_any_user(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-012: Admin can create token for any user."""
    admin = admin_test_resources["admin"]
    target_user = "bob@example.com"

    with http_client_factory(admin) as client:
        api = "api/2.0/mlflow/users/access-token"
        resp = client.patch(api, json={"username": target_user})

        if resp.status_code == 404:
            pytest.skip("Access token endpoint not available")

        assert resp.status_code in (200, 201), f"Admin should create token: {resp.status_code}"

        data = resp.json()
        assert "token" in data, "Response should contain token"


@pytest.mark.integration
def test_admin_creates_token_for_service_account(
    http_client_factory,
    admin_test_resources,
    test_run_id: str,
) -> None:
    """ADM-013: Admin can create token for service account."""
    admin = admin_test_resources["admin"]
    svc_username = f"svc-token-admin-{test_run_id}@example.com"

    with http_client_factory(admin) as client:
        # Create service account first
        create_api = "api/2.0/mlflow/users"
        create_resp = client.post(
            create_api,
            json={
                "username": svc_username,
                "display_name": f"Token Test {test_run_id}",
                "is_admin": False,
                "is_service_account": True,
            },
        )
        assert create_resp.status_code in (200, 201), "Failed to create service account"

        # Create token
        token_api = "api/2.0/mlflow/users/access-token"
        token_resp = client.patch(token_api, json={"username": svc_username})

        if token_resp.status_code == 404:
            pytest.skip("Access token endpoint not available")

        assert token_resp.status_code in (200, 201), f"Admin should create token for svc account: {token_resp.status_code}"


@pytest.mark.integration
def test_admin_updates_user_attributes(
    http_client_factory,
    admin_test_resources,
    test_run_id: str,
) -> None:
    """ADM-014: Admin can update user attributes."""
    admin = admin_test_resources["admin"]
    target_user = f"update-test-{test_run_id}@example.com"

    with http_client_factory(admin) as client:
        # Create user first
        create_api = "api/2.0/mlflow/users"
        client.post(
            create_api,
            json={
                "username": target_user,
                "display_name": f"Update Test {test_run_id}",
                "is_admin": False,
                "is_service_account": False,
            },
        )

        # Update user (endpoint may vary)
        update_api = "api/2.0/mlflow/users"
        resp = client.patch(
            update_api,
            json={
                "username": target_user,
                "display_name": f"Updated Display Name {test_run_id}",
            },
        )

        # May be 200, 404 if endpoint doesn't exist, or 501 if not implemented
        assert resp.status_code in (200, 201, 404, 501), f"Unexpected status: {resp.status_code}"


# =============================================================================
# ADM-020 to ADM-022: Admin Bypasses All Restrictions
# =============================================================================


@pytest.mark.integration
def test_admin_accesses_resource_with_no_permissions(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-020: Admin accesses resource even with NO_PERMISSIONS granted."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        # Admin was granted NO_PERMISSIONS on restricted_exp, but should still access
        api = f"ajax-api/2.0/mlflow/experiments/get-by-name?experiment_name={quote(resources['restricted_exp_name'])}"
        resp = client.get(api)
        assert _is_ok(resp.status_code), f"Admin should bypass NO_PERMISSIONS: {resp.status_code}"


@pytest.mark.integration
def test_admin_modifies_read_only_resource(
    http_client_factory,
    admin_test_resources,
) -> None:
    """ADM-021: Admin modifies resource even with only READ permission."""
    resources = admin_test_resources
    admin = resources["admin"]

    with http_client_factory(admin) as client:
        # Admin was granted READ on experiment, but should still create runs
        api = "api/2.0/mlflow/runs/create"
        resp = client.post(
            api,
            json={
                "experiment_id": resources["experiment_id"],
                "start_time": int(time.time() * 1000),
            },
        )
        assert _is_ok(resp.status_code), f"Admin should bypass READ-only: {resp.status_code}"


@pytest.mark.integration
def test_admin_deletes_resource_without_manage(
    http_client_factory,
    admin_test_resources,
    test_run_id: str,
) -> None:
    """ADM-022: Admin deletes resource even without MANAGE permission."""
    resources = admin_test_resources
    admin = resources["admin"]
    owner = resources["owner"]

    # Create a new model for delete test
    delete_model_name = f"admin-delete-model-{test_run_id}-{uuid.uuid4().hex[:6]}"

    with http_client_factory(owner) as client:
        assert _create_model(client, delete_model_name), "Failed to create model"
        # Grant admin EDIT (not MANAGE)
        _grant_user_permission(client, "model", delete_model_name, admin, "EDIT")

    with http_client_factory(admin) as client:
        # Admin should still be able to delete despite only having EDIT
        api = "ajax-api/2.0/mlflow/registered-models/delete"
        resp = client.request("DELETE", api, json={"name": delete_model_name})
        assert _is_ok(resp.status_code), f"Admin should bypass EDIT-only for delete: {resp.status_code}"



from __future__ import annotations

from typing import Any

from mlflow_oidc_auth.graphql.patch import (
    install_mlflow_graphql_authorization_middleware,
    uninstall_mlflow_graphql_authorization_middleware,
)


def test_patch_installs_into_mlflow_handlers() -> None:
    """Ensure install patches MLflow's handler hook used by /graphql."""

    import mlflow.server.handlers as mlflow_handlers

    # The application factory may install the patch at import time.
    # Normalize to a clean baseline for this test.
    uninstall_mlflow_graphql_authorization_middleware()

    original = getattr(mlflow_handlers, "_get_graphql_auth_middleware")
    try:
        install_mlflow_graphql_authorization_middleware()
        middleware = mlflow_handlers._get_graphql_auth_middleware()  # type: ignore[attr-defined]
        assert isinstance(middleware, list)
        assert len(middleware) == 1
        # Don't assert exact type path too strictly; just ensure it's callable middleware.
        assert hasattr(middleware[0], "resolve")
    finally:
        uninstall_mlflow_graphql_authorization_middleware()
        assert getattr(mlflow_handlers, "_get_graphql_auth_middleware") is original



from typing import List, Optional, Callable
from sqlalchemy.orm import Session

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import RESOURCE_DOES_NOT_EXIST

from mlflow_oidc_auth.db.models import SqlGroup, SqlRegisteredModelGroupPermission
from mlflow_oidc_auth.entities import RegisteredModelPermission
from mlflow_oidc_auth.permissions import _validate_permission, compare_permissions
from mlflow_oidc_auth.repository import GroupRepository
from mlflow_oidc_auth.repository.utils import get_user, get_group, list_user_groups


class RegisteredModelPermissionGroupRepository:
    def __init__(self, session_maker):
        self._Session: Callable[[], Session] = session_maker
        self._group_repo = GroupRepository(session_maker)

    def _get_registered_model_group_permission(self, session: Session, name: str, group_name: str) -> Optional[SqlRegisteredModelGroupPermission]:
        group = session.query(SqlGroup).filter(SqlGroup.group_name == group_name).one_or_none()
        if group is None:
            return None
        return (
            session.query(SqlRegisteredModelGroupPermission)
            .filter(
                SqlRegisteredModelGroupPermission.name == name,
                SqlRegisteredModelGroupPermission.group_id == group.id,
            )
            .one_or_none()
        )

    def create(self, group_name: str, name: str, permission: str):
        _validate_permission(permission)
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = SqlRegisteredModelGroupPermission(name=name, group_id=group.id, permission=permission)
            session.add(perm)
            session.flush()
            return perm.to_mlflow_entity()

    def get(self, group_name: str) -> List[RegisteredModelPermission]:
        with self._Session() as session:
            group = get_group(session, group_name)
            perms = session.query(SqlRegisteredModelGroupPermission).filter(SqlRegisteredModelGroupPermission.group_id == group.id).all()
            return [p.to_mlflow_entity() for p in perms]

    def list_groups_for_model(self, name: str) -> List[tuple[str, str]]:
        """List groups that have explicit permissions for a registered model.

        Returns pairs of (group_name, permission).
        """

        with self._Session() as session:
            rows = (
                session.query(SqlGroup.group_name, SqlRegisteredModelGroupPermission.permission)
                .join(SqlRegisteredModelGroupPermission, SqlRegisteredModelGroupPermission.group_id == SqlGroup.id)
                .filter(SqlRegisteredModelGroupPermission.name == name)
                .filter(SqlRegisteredModelGroupPermission.prompt == False)
                .all()
            )
            return [(str(group_name), str(permission)) for group_name, permission in rows]

    def get_for_user(self, name: str, username: str) -> RegisteredModelPermission:
        with self._Session() as session:
            user_groups = self._group_repo.list_groups_for_user(username)
            user_perms: Optional[SqlRegisteredModelGroupPermission] = None
            for ug in user_groups:
                perms = self._get_registered_model_group_permission(session, name, ug)
                if perms is None:
                    continue
                if user_perms is None:
                    user_perms = perms
                    continue
                try:
                    if compare_permissions(str(user_perms.permission), str(perms.permission)):
                        user_perms = perms
                except AttributeError:
                    user_perms = perms
            try:
                if user_perms is not None:
                    return user_perms.to_mlflow_entity()
                else:
                    raise MlflowException(
                        f"Registered model permission with name={name} and username={username} not found",
                        RESOURCE_DOES_NOT_EXIST,
                    )
            except AttributeError:
                raise MlflowException(
                    f"Registered model permission with name={name} and username={username} not found",
                    RESOURCE_DOES_NOT_EXIST,
                )

    def list_for_user(self, username: str) -> List[RegisteredModelPermission]:
        with self._Session() as session:
            user = get_user(session, username=username)
            user_groups = list_user_groups(session, user)
            perms = (
                session.query(SqlRegisteredModelGroupPermission)
                .filter(SqlRegisteredModelGroupPermission.group_id.in_([ug.group_id for ug in user_groups]))
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update(self, group_name: str, name: str, permission: str):
        _validate_permission(permission)
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = (
                session.query(SqlRegisteredModelGroupPermission)
                .filter(SqlRegisteredModelGroupPermission.name == name, SqlRegisteredModelGroupPermission.group_id == group.id)
                .one()
            )
            perm.permission = permission
            session.flush()
            return perm.to_mlflow_entity()

    def rename(self, old_name: str, new_name: str):
        with self._Session() as session:
            perms = session.query(SqlRegisteredModelGroupPermission).filter(SqlRegisteredModelGroupPermission.name == old_name).all()
            if not perms:
                raise MlflowException(f"No registered model group permissions found for name: {old_name}", RESOURCE_DOES_NOT_EXIST)
            for perm in perms:
                perm.name = new_name
            session.flush()

    def delete(self, group_name: str, name: str):
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = (
                session.query(SqlRegisteredModelGroupPermission)
                .filter(SqlRegisteredModelGroupPermission.name == name, SqlRegisteredModelGroupPermission.group_id == group.id)
                .one()
            )
            session.delete(perm)
            session.flush()

    def wipe(self, name: str):
        with self._Session() as session:
            perms = session.query(SqlRegisteredModelGroupPermission).filter(SqlRegisteredModelGroupPermission.name == name).all()
            for p in perms:
                session.delete(p)
            session.flush()



"""
Unit tests for configuration providers.

Tests the pluggable configuration provider system including:
- Base provider interface
- Environment variable provider
- Config manager chain of responsibility
- Provider priority and fallback behavior
"""

import os
import tempfile
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from mlflow_oidc_auth.config_providers.base import ConfigProvider, SecretLevel, get_secret_level
from mlflow_oidc_auth.config_providers.env_provider import EnvProvider
from mlflow_oidc_auth.config_providers.manager import ConfigManager


class TestSecretLevel(unittest.TestCase):
    """Tests for secret level classification."""

    def test_secret_keys_classified_as_secret(self) -> None:
        """Verify that sensitive keys are classified as SECRET."""
        self.assertEqual(get_secret_level("SECRET_KEY"), SecretLevel.SECRET)
        self.assertEqual(get_secret_level("OIDC_CLIENT_SECRET"), SecretLevel.SECRET)

    def test_sensitive_keys_classified_as_sensitive(self) -> None:
        """Verify that sensitive config keys are classified correctly."""
        self.assertEqual(get_secret_level("OIDC_USERS_DB_URI"), SecretLevel.SENSITIVE)
        self.assertEqual(get_secret_level("OIDC_CLIENT_ID"), SecretLevel.SENSITIVE)

    def test_unknown_keys_classified_as_public(self) -> None:
        """Verify that unknown keys default to PUBLIC."""
        self.assertEqual(get_secret_level("UNKNOWN_KEY"), SecretLevel.PUBLIC)
        self.assertEqual(get_secret_level("EXTEND_MLFLOW_MENU"), SecretLevel.PUBLIC)


class TestEnvProvider(unittest.TestCase):
    """Tests for the environment variable provider."""

    def setUp(self) -> None:
        self.provider = EnvProvider()

    def test_name(self) -> None:
        """Verify provider name."""
        self.assertEqual(self.provider.name, "env")

    def test_priority(self) -> None:
        """Verify env provider has lowest priority."""
        self.assertEqual(self.provider.priority, 1000)

    def test_is_available(self) -> None:
        """Env provider is always available."""
        self.assertTrue(self.provider.is_available())

    def test_get_existing_variable(self) -> None:
        """Get an existing environment variable."""
        with patch.dict(os.environ, {"TEST_VAR": "test_value"}):
            self.assertEqual(self.provider.get("TEST_VAR"), "test_value")

    def test_get_missing_variable_returns_default(self) -> None:
        """Missing variable returns default."""
        self.assertEqual(self.provider.get("NONEXISTENT_VAR", "default"), "default")

    def test_get_missing_variable_returns_none(self) -> None:
        """Missing variable without default returns None."""
        self.assertIsNone(self.provider.get("NONEXISTENT_VAR"))

    def test_get_many(self) -> None:
        """Get multiple variables at once."""
        with patch.dict(os.environ, {"VAR1": "value1", "VAR2": "value2"}):
            result = self.provider.get_many(["VAR1", "VAR2", "VAR3"])
            self.assertEqual(result, {"VAR1": "value1", "VAR2": "value2"})


class MockSecretProvider(ConfigProvider):
    """Mock provider for testing that handles secrets."""

    def __init__(self, secrets: dict) -> None:
        self._secrets = secrets

    @property
    def name(self) -> str:
        return "mock-secrets"

    @property
    def priority(self) -> int:
        return 10

    def is_available(self) -> bool:
        return True

    def get(self, key: str, default=None):
        level = get_secret_level(key)
        if level == SecretLevel.PUBLIC:
            return default
        return self._secrets.get(key, default)


class MockConfigProvider(ConfigProvider):
    """Mock provider for testing that handles config values."""

    def __init__(self, config: dict, priority: int = 50) -> None:
        self._config = config
        self._priority = priority

    @property
    def name(self) -> str:
        return "mock-config"

    @property
    def priority(self) -> int:
        return self._priority

    def is_available(self) -> bool:
        return True

    def get(self, key: str, default=None):
        return self._config.get(key, default)


class TestConfigManager(unittest.TestCase):
    """Tests for the ConfigManager."""

    def test_provider_priority_order(self) -> None:
        """Verify providers are sorted by priority."""
        manager = ConfigManager(auto_discover=False)

        provider1 = MockConfigProvider({}, priority=100)
        provider2 = MockSecretProvider({})
        # MockSecretProvider has priority 10

        manager.register_provider(provider1)
        manager.register_provider(provider2)

        # Lower priority number = higher priority
        self.assertEqual(manager.providers[0].name, "mock-secrets")
        self.assertEqual(manager.providers[1].name, "mock-config")

    def test_chain_of_responsibility(self) -> None:
        """First provider with value wins."""
        manager = ConfigManager(auto_discover=False)

        secret_provider = MockSecretProvider({"OIDC_CLIENT_SECRET": "secret_value"})
        config_provider = MockConfigProvider({"SOME_CONFIG": "config_value", "OIDC_CLIENT_SECRET": "wrong_value"})
        env_provider = EnvProvider()

        manager.register_provider(secret_provider)
        manager.register_provider(config_provider)
        manager.register_provider(env_provider)

        # Secret comes from secret provider
        self.assertEqual(manager.get("OIDC_CLIENT_SECRET"), "secret_value")
        # Config comes from config provider
        self.assertEqual(manager.get("SOME_CONFIG"), "config_value")

    def test_fallback_to_default(self) -> None:
        """Returns default when no provider has the value."""
        manager = ConfigManager(auto_discover=False)
        manager.register_provider(EnvProvider())

        self.assertEqual(manager.get("NONEXISTENT", "default_value"), "default_value")

    def test_get_bool(self) -> None:
        """Test boolean value parsing."""
        manager = ConfigManager(auto_discover=False)
        manager.register_provider(MockConfigProvider({"TRUE_VAL": "true", "FALSE_VAL": "false", "ONE_VAL": "1", "T_VAL": "T"}))

        self.assertTrue(manager.get_bool("TRUE_VAL"))
        self.assertFalse(manager.get_bool("FALSE_VAL"))
        self.assertTrue(manager.get_bool("ONE_VAL"))
        self.assertTrue(manager.get_bool("T_VAL"))
        self.assertFalse(manager.get_bool("NONEXISTENT"))
        self.assertTrue(manager.get_bool("NONEXISTENT", default=True))

    def test_get_int(self) -> None:
        """Test integer value parsing."""
        manager = ConfigManager(auto_discover=False)
        manager.register_provider(MockConfigProvider({"NUM": "42", "INVALID": "not_a_number"}))

        self.assertEqual(manager.get_int("NUM"), 42)
        self.assertEqual(manager.get_int("INVALID", default=0), 0)
        self.assertEqual(manager.get_int("NONEXISTENT", default=100), 100)

    def test_get_list(self) -> None:
        """Test list value parsing."""
        manager = ConfigManager(auto_discover=False)
        manager.register_provider(MockConfigProvider({"LIST": "a, b, c", "SINGLE": "only_one"}))

        self.assertEqual(manager.get_list("LIST"), ["a", "b", "c"])
        self.assertEqual(manager.get_list("SINGLE"), ["only_one"])
        self.assertEqual(manager.get_list("NONEXISTENT", default=["default"]), ["default"])

    def test_refresh_calls_all_providers(self) -> None:
        """Refresh should call refresh on all providers."""
        manager = ConfigManager(auto_discover=False)

        mock_provider = MockConfigProvider({})
        mock_provider.refresh = MagicMock()

        manager.register_provider(mock_provider)
        manager.refresh()

        mock_provider.refresh.assert_called_once()

    def test_close_calls_all_providers(self) -> None:
        """Close should call close on all providers."""
        manager = ConfigManager(auto_discover=False)

        mock_provider = MockConfigProvider({})
        mock_provider.close = MagicMock()

        manager.register_provider(mock_provider)
        manager.close()

        mock_provider.close.assert_called_once()


class TestKubernetesSecretsProvider(unittest.TestCase):
    """Tests for Kubernetes Secrets provider."""

    def test_reads_secrets_from_mounted_files(self) -> None:
        """Verify secrets are read from mounted files."""
        from mlflow_oidc_auth.config_providers.kubernetes_provider import KubernetesSecretsProvider

        with tempfile.TemporaryDirectory() as tmpdir:
            # Create mock secret files
            secret_path = Path(tmpdir)
            (secret_path / "OIDC_CLIENT_SECRET").write_text("my_secret_value")
            (secret_path / "SECRET_KEY").write_text("session_key\n")  # With trailing newline

            with patch.dict(os.environ, {"CONFIG_K8S_SECRETS_ENABLED": "true", "CONFIG_K8S_SECRETS_PATH": tmpdir}):
                provider = KubernetesSecretsProvider()
                self.assertTrue(provider.is_available())

                # Should read and strip newlines
                self.assertEqual(provider.get("OIDC_CLIENT_SECRET"), "my_secret_value")
                self.assertEqual(provider.get("SECRET_KEY"), "session_key")

                # Public keys should return default
                self.assertIsNone(provider.get("EXTEND_MLFLOW_MENU"))

    def test_not_available_when_disabled(self) -> None:
        """Provider is not available when not enabled."""
        from mlflow_oidc_auth.config_providers.kubernetes_provider import KubernetesSecretsProvider

        with patch.dict(os.environ, {"CONFIG_K8S_SECRETS_ENABLED": "false"}, clear=False):
            provider = KubernetesSecretsProvider()
            self.assertFalse(provider.is_available())


class TestAWSProvidersAvailability(unittest.TestCase):
    """Tests for AWS provider availability checks."""

    def test_secrets_manager_not_available_when_disabled(self) -> None:
        """AWS Secrets Manager is not available when not enabled."""
        from mlflow_oidc_auth.config_providers.aws_secrets_provider import AWSSecretsManagerProvider

        with patch.dict(os.environ, {"CONFIG_AWS_SECRETS_ENABLED": "false"}, clear=False):
            provider = AWSSecretsManagerProvider()
            self.assertFalse(provider.is_available())

    def test_parameter_store_not_available_when_disabled(self) -> None:
        """AWS Parameter Store is not available when not enabled."""
        from mlflow_oidc_auth.config_providers.aws_parameter_store_provider import AWSParameterStoreProvider

        with patch.dict(os.environ, {"CONFIG_AWS_PARAMETER_STORE_ENABLED": "false"}, clear=False):
            provider = AWSParameterStoreProvider()
            self.assertFalse(provider.is_available())


class TestAzureProviderAvailability(unittest.TestCase):
    """Tests for Azure provider availability checks."""

    def test_keyvault_not_available_when_disabled(self) -> None:
        """Azure Key Vault is not available when not enabled."""
        from mlflow_oidc_auth.config_providers.azure_keyvault_provider import AzureKeyVaultProvider

        with patch.dict(os.environ, {"CONFIG_AZURE_KEYVAULT_ENABLED": "false"}, clear=False):
            provider = AzureKeyVaultProvider()
            self.assertFalse(provider.is_available())

    def test_keyvault_not_available_without_vault_name(self) -> None:
        """Azure Key Vault requires vault name."""
        from mlflow_oidc_auth.config_providers.azure_keyvault_provider import AzureKeyVaultProvider

        with patch.dict(os.environ, {"CONFIG_AZURE_KEYVAULT_ENABLED": "true", "CONFIG_AZURE_KEYVAULT_NAME": ""}, clear=False):
            provider = AzureKeyVaultProvider()
            self.assertFalse(provider.is_available())


class TestVaultProviderAvailability(unittest.TestCase):
    """Tests for HashiCorp Vault provider availability checks."""

    def test_vault_not_available_when_disabled(self) -> None:
        """HashiCorp Vault is not available when not enabled."""
        from mlflow_oidc_auth.config_providers.vault_provider import HashiCorpVaultProvider

        with patch.dict(os.environ, {"CONFIG_VAULT_ENABLED": "false"}, clear=False):
            provider = HashiCorpVaultProvider()
            self.assertFalse(provider.is_available())

    def test_vault_not_available_without_auth(self) -> None:
        """HashiCorp Vault requires authentication credentials."""
        from mlflow_oidc_auth.config_providers.vault_provider import HashiCorpVaultProvider

        with patch.dict(
            os.environ, {"CONFIG_VAULT_ENABLED": "true", "CONFIG_VAULT_TOKEN": "", "CONFIG_VAULT_ROLE_ID": "", "CONFIG_VAULT_SECRET_ID": ""}, clear=False
        ):
            provider = HashiCorpVaultProvider()
            self.assertFalse(provider.is_available())


if __name__ == "__main__":
    unittest.main()



"""
Comprehensive tests for the registered model permissions router.

This module tests all registered model permission endpoints including listing models,
getting model user permissions with various scenarios including authentication,
authorization, and error handling.
"""

import pytest
from unittest.mock import MagicMock, patch

from mlflow_oidc_auth.routers.registered_model_permissions import (
    registered_model_permissions_router,
    get_registered_model_groups,
    get_registered_model_users,
    list_models,
    LIST_MODELS,
    REGISTERED_MODEL_GROUP_PERMISSIONS,
    REGISTERED_MODEL_USER_PERMISSIONS,
)
from mlflow_oidc_auth.entities import User, RegisteredModelPermission as RegisteredModelPermissionEntity


class TestRegisteredModelPermissionsRouter:
    """Test class for registered model permissions router configuration."""

    def test_router_configuration(self):
        """Test that the registered model permissions router is properly configured."""
        assert registered_model_permissions_router.prefix == "/api/2.0/mlflow/permissions/registered-models"
        assert "registered model permissions" in registered_model_permissions_router.tags
        assert 403 in registered_model_permissions_router.responses
        assert 404 in registered_model_permissions_router.responses

    def test_route_constants(self):
        """Test that route constants are properly defined."""
        assert LIST_MODELS == ""
        assert REGISTERED_MODEL_USER_PERMISSIONS == "/{name}/users"
        assert REGISTERED_MODEL_GROUP_PERMISSIONS == "/{name}/groups"


class TestGetRegisteredModelGroupsEndpoint:
    """Test the get registered model groups endpoint functionality."""

    @pytest.mark.asyncio
    async def test_get_registered_model_groups_success(self, mock_store):
        mock_store.registered_model_group_repo.list_groups_for_model.return_value = [("team-a", "READ"), ("team-b", "MANAGE")]

        with patch("mlflow_oidc_auth.routers.registered_model_permissions.store", mock_store):
            result = await get_registered_model_groups(name="test-model", _="admin@example.com")

        assert len(result) == 2
        assert result[0].name == "team-a"
        assert result[0].permission == "READ"
        assert result[0].kind == "group"
        assert result[1].name == "team-b"
        assert result[1].permission == "MANAGE"
        assert result[1].kind == "group"

    def test_get_registered_model_groups_integration(self, admin_client, mock_store):
        mock_store.registered_model_group_repo.list_groups_for_model.return_value = []
        response = admin_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/groups")
        assert response.status_code == 200


class TestGetRegisteredModelUsersEndpoint:
    """Test the get registered model users endpoint functionality."""

    @pytest.mark.asyncio
    async def test_get_registered_model_users_success(self, mock_store):
        """Test successful retrieval of registered model users."""
        # Mock users with registered model permissions
        user1 = User(
            id_=1,
            username="user1@example.com",
            password_hash="hash1",
            password_expiration=None,
            display_name="User 1",
            is_admin=False,
            is_service_account=False,
            registered_model_permissions=[RegisteredModelPermissionEntity(name="test-model", permission="MANAGE")],
        )

        user2 = User(
            id_=2,
            username="service@example.com",
            password_hash="hash2",
            password_expiration=None,
            display_name="Service Account",
            is_admin=False,
            is_service_account=True,
            registered_model_permissions=[RegisteredModelPermissionEntity(name="test-model", permission="READ")],
        )

        user3 = User(
            id_=3,
            username="user3@example.com",
            password_hash="hash3",
            password_expiration=None,
            display_name="User 3",
            is_admin=False,
            is_service_account=False,
            registered_model_permissions=[],  # No permissions for this model
        )

        mock_store.list_users.return_value = [user1, user2, user3]

        with patch("mlflow_oidc_auth.routers.registered_model_permissions.store", mock_store):
            result = await get_registered_model_users(name="test-model", _=None)

        assert len(result) == 2  # Only users with permissions for test-model

        # Check first user
        assert result[0].name == "user1@example.com"
        assert result[0].permission == "MANAGE"
        assert result[0].kind == "user"

        # Check service account
        assert result[1].name == "service@example.com"
        assert result[1].permission == "READ"
        assert result[1].kind == "service-account"

    @pytest.mark.asyncio
    async def test_get_registered_model_users_no_permissions(self, mock_store):
        """Test getting registered model users when no users have permissions."""
        user1 = User(username="user1@example.com", display_name="User 1", is_admin=False, is_service_account=False, registered_model_permissions=[])

        mock_store.list_users.return_value = [user1]

        result = await get_registered_model_users(name="test-model", _=None)

        assert len(result) == 0

    @pytest.mark.asyncio
    async def test_get_registered_model_users_multiple_models(self, mock_store):
        """Test getting users for specific model when users have multiple model permissions."""
        user1 = User(
            username="user1@example.com",
            display_name="User 1",
            is_admin=False,
            is_service_account=False,
            registered_model_permissions=[
                RegisteredModelPermissionEntity(name="model-1", permission="MANAGE"),
                RegisteredModelPermissionEntity(name="model-2", permission="READ"),
            ],
        )

        mock_store.list_users.return_value = [user1]

        result = await get_registered_model_users(name="model-1", _=None)

        assert len(result) == 1
        assert result[0].name == "user1@example.com"
        assert result[0].permission == "MANAGE"  # Should get permission for model-1

    @pytest.mark.asyncio
    async def test_get_registered_model_users_no_registered_model_permissions_attr(self, mock_store):
        """Test getting users when user object doesn't have registered_model_permissions attribute."""
        user1 = User(username="user1@example.com", display_name="User 1", is_admin=False, is_service_account=False)
        # Remove the registered_model_permissions attribute
        delattr(user1, "registered_model_permissions")

        mock_store.list_users.return_value = [user1]

        result = await get_registered_model_users(name="test-model", _=None)

        assert len(result) == 0

    def test_get_registered_model_users_integration(self, admin_client):
        """Test get registered model users endpoint through FastAPI test client."""
        response = admin_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")

        assert response.status_code == 200
        assert isinstance(response.json(), list)

    def test_get_registered_model_users_non_admin(self, authenticated_client):
        """Test get registered model users as non-admin user."""
        response = authenticated_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")

        # Allowed because mocks grant manage permission
        assert response.status_code == 200

    def test_get_registered_model_users_non_admin_without_manage_permission(self, authenticated_client):
        """Non-admin lacking manage rights should be forbidden."""
        with patch("mlflow_oidc_auth.dependencies.can_manage_registered_model", return_value=False):
            response = authenticated_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")

        assert response.status_code == 403

    def test_get_registered_model_users_unauthenticated(self, client):
        """Test get registered model users without authentication."""
        response = client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")

        # Should fail due to authentication requirement
        assert response.status_code in [401, 403]


class TestListModelsEndpoint:
    """Test the list models endpoint functionality."""

    @pytest.mark.asyncio
    async def test_list_models_admin(self):
        """Test listing models as admin user."""
        # Mock registered models
        mock_model1 = MagicMock()
        mock_model1.name = "model-1"
        mock_model1.tags = {"env": "test"}
        mock_model1.description = "Test Model 1"
        mock_model1.aliases = ["alias1"]

        mock_model2 = MagicMock()
        mock_model2.name = "model-2"
        mock_model2.tags = {"env": "prod"}
        mock_model2.description = "Test Model 2"
        mock_model2.aliases = []

        with patch("mlflow_oidc_auth.routers.registered_model_permissions.fetch_all_registered_models") as mock_fetch:
            mock_fetch.return_value = [mock_model1, mock_model2]

            result = await list_models(username="admin@example.com", is_admin=True)

            assert result.status_code == 200

            import json

            content = json.loads(result.body.decode())

            assert len(content) == 2
            assert content[0]["name"] == "model-1"
            assert content[0]["tags"] == {"env": "test"}
            assert content[0]["description"] == "Test Model 1"
            assert content[0]["aliases"] == ["alias1"]

    @pytest.mark.asyncio
    async def test_list_models_regular_user_with_permissions(self):
        """Test listing models as regular user with manage permissions."""
        mock_model1 = MagicMock()
        mock_model1.name = "model-1"
        mock_model1.tags = {"env": "test"}
        mock_model1.description = "Test Model 1"
        mock_model1.aliases = []

        mock_model2 = MagicMock()
        mock_model2.name = "model-2"
        mock_model2.tags = {"env": "prod"}
        mock_model2.description = "Test Model 2"
        mock_model2.aliases = []

        # Mock filter_manageable_models to return only model-1
        with patch("mlflow_oidc_auth.routers.registered_model_permissions.fetch_all_registered_models") as mock_fetch, patch(
            "mlflow_oidc_auth.routers.registered_model_permissions.filter_manageable_models"
        ) as mock_filter:
            mock_fetch.return_value = [mock_model1, mock_model2]
            mock_filter.return_value = [mock_model1]  # Only model-1 is manageable

            result = await list_models(username="user@example.com", is_admin=False)

            assert result.status_code == 200

            import json

            content = json.loads(result.body.decode())

            assert len(content) == 1  # Only model-1 should be returned
            assert content[0]["name"] == "model-1"

            # Verify filter_manageable_models was called with correct args
            mock_filter.assert_called_once_with("user@example.com", [mock_model1, mock_model2])

    @pytest.mark.asyncio
    async def test_list_models_regular_user_no_permissions(self):
        """Test listing models as regular user with no permissions."""
        mock_model1 = MagicMock()
        mock_model1.name = "model-1"
        mock_model1.tags = {}
        mock_model1.description = "Test Model 1"
        mock_model1.aliases = []

        with patch("mlflow_oidc_auth.routers.registered_model_permissions.fetch_all_registered_models") as mock_fetch, patch(
            "mlflow_oidc_auth.routers.registered_model_permissions.filter_manageable_models", return_value=[]
        ):
            mock_fetch.return_value = [mock_model1]

            result = await list_models(username="user@example.com", is_admin=False)

            assert result.status_code == 200

            import json

            content = json.loads(result.body.decode())

            assert len(content) == 0

    @pytest.mark.asyncio
    async def test_list_models_empty_list(self):
        """Test listing models when no models exist."""
        with patch("mlflow_oidc_auth.routers.registered_model_permissions.fetch_all_registered_models") as mock_fetch:
            mock_fetch.return_value = []

            result = await list_models(username="admin@example.com", is_admin=True)

            assert result.status_code == 200

            import json

            content = json.loads(result.body.decode())

            assert len(content) == 0

    @pytest.mark.asyncio
    async def test_list_models_with_none_values(self):
        """Test listing models with None values in model attributes."""
        mock_model = MagicMock()
        mock_model.name = "model-1"
        mock_model.tags = None
        mock_model.description = None
        mock_model.aliases = None

        with patch("mlflow_oidc_auth.routers.registered_model_permissions.fetch_all_registered_models") as mock_fetch:
            mock_fetch.return_value = [mock_model]

            result = await list_models(username="admin@example.com", is_admin=True)

            assert result.status_code == 200

            import json

            content = json.loads(result.body.decode())

            assert len(content) == 1
            assert content[0]["name"] == "model-1"
            assert content[0]["tags"] is None
            assert content[0]["description"] is None
            assert content[0]["aliases"] is None

    def test_list_models_integration_admin(self, admin_client):
        """Test list models endpoint through FastAPI test client as admin."""
        response = admin_client.get("/api/2.0/mlflow/permissions/registered-models")

        assert response.status_code == 200
        assert isinstance(response.json(), list)

    def test_list_models_integration_regular_user(self, authenticated_client):
        """Test list models endpoint through FastAPI test client as regular user."""
        response = authenticated_client.get("/api/2.0/mlflow/permissions/registered-models")

        assert response.status_code == 200
        assert isinstance(response.json(), list)

    def test_list_models_unauthenticated(self, client):
        """Test list models without authentication."""
        response = client.get("/api/2.0/mlflow/permissions/registered-models")

        # Should fail due to authentication requirement
        assert response.status_code in [401, 403]


class TestRegisteredModelPermissionsRouterIntegration:
    """Test class for registered model permissions router integration scenarios."""

    def test_all_endpoints_require_authentication(self, client):
        """Test that all registered model permission endpoints require authentication."""
        endpoints = [("GET", "/api/2.0/mlflow/permissions/registered-models"), ("GET", "/api/2.0/mlflow/permissions/registered-models/test-model/users")]

        for method, endpoint in endpoints:
            response = client.get(endpoint)

            # Should require authentication
            assert response.status_code in [401, 403]

    def test_model_user_permissions_requires_admin(self, authenticated_client):
        """Test that model user permissions endpoint requires admin privileges."""
        with patch("mlflow_oidc_auth.dependencies.can_manage_registered_model", return_value=False):
            response = authenticated_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")

        # Without manage permission a non-admin is forbidden
        assert response.status_code == 403

    def test_endpoints_response_content_type(self, authenticated_client, admin_client):
        """Test that endpoints return proper content type."""
        # Test list models
        response = authenticated_client.get("/api/2.0/mlflow/permissions/registered-models")
        assert "application/json" in response.headers.get("content-type", "")

        # Test model users (admin only)
        response = admin_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")
        assert "application/json" in response.headers.get("content-type", "")

    def test_model_name_parameter_validation(self, admin_client):
        """Test model name parameter validation."""
        # Test with various model name formats
        model_names = ["test-model", "model_with_underscores", "model123", "Model-Name"]

        for model_name in model_names:
            response = admin_client.get(f"/api/2.0/mlflow/permissions/registered-models/{model_name}/users")

            # Should not fail due to parameter format
            assert response.status_code == 200

    def test_registered_model_permissions_response_structure(self, authenticated_client):
        """Test that registered model permissions endpoints return proper response structure."""
        # Test list models response structure
        response = authenticated_client.get("/api/2.0/mlflow/permissions/registered-models")

        assert response.status_code == 200
        models = response.json()
        assert isinstance(models, list)

        if models:  # If there are models
            model = models[0]
            assert "name" in model
            assert "tags" in model
            assert "description" in model
            assert "aliases" in model

    def test_model_users_response_structure(self, admin_client):
        """Test that model users endpoint returns proper response structure."""
        response = admin_client.get("/api/2.0/mlflow/permissions/registered-models/test-model/users")

        assert response.status_code == 200
        users = response.json()
        assert isinstance(users, list)

        if users:  # If there are users with permissions
            user = users[0]
            assert "name" in user
            assert "permission" in user
            assert "kind" in user
            assert user["kind"] in ["user", "service-account"]

    def test_registered_model_permissions_error_handling(self, admin_client):
        """Test error handling in registered model permissions endpoints."""
        # Test with empty model name
        response = admin_client.get("/api/2.0/mlflow/permissions/registered-models//users")

        # Should handle invalid paths gracefully
        assert response.status_code in [404, 422]

    def test_registered_model_permissions_concurrent_requests(self, authenticated_client):
        """Test that registered model permissions endpoints handle concurrent requests."""
        import concurrent.futures

        def make_request():
            return authenticated_client.get("/api/2.0/mlflow/permissions/registered-models")

        # Make concurrent requests
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            futures = [executor.submit(make_request) for _ in range(5)]

            for future in concurrent.futures.as_completed(futures):
                response = future.result()
                assert response.status_code == 200  # Should not crash

    def test_registered_model_permissions_with_special_characters(self, admin_client):
        """Test registered model permissions with special characters in model name."""
        # Test with URL-encoded special characters
        special_names = ["model%20name", "model-with-dashes", "model_with_underscores"]

        for model_name in special_names:
            response = admin_client.get(f"/api/2.0/mlflow/permissions/registered-models/{model_name}/users")

            # Should handle special characters (may return empty list, but shouldn't crash)
            assert response.status_code == 200

    def test_registered_model_permissions_performance(self, authenticated_client):
        """Test that registered model permissions endpoints respond in reasonable time."""
        import time

        endpoints = ["/api/2.0/mlflow/permissions/registered-models"]

        for endpoint in endpoints:
            start_time = time.time()
            response = authenticated_client.get(endpoint)
            end_time = time.time()

            # Should respond within reasonable time (5 seconds)
            assert (end_time - start_time) < 5.0
            assert response.status_code == 200

    def test_registered_model_permissions_with_long_model_names(self, admin_client):
        """Test registered model permissions with very long model names."""
        # Test with a very long model name
        long_model_name = "a" * 1000  # 1000 character model name

        response = admin_client.get(f"/api/2.0/mlflow/permissions/registered-models/{long_model_name}/users")

        # Should handle long names gracefully (may return 404 or empty list)
        assert response.status_code in [200, 404]

    def test_registered_model_permissions_case_sensitivity(self, admin_client):
        """Test registered model permissions case sensitivity."""
        # Test with different cases
        model_names = ["TestModel", "testmodel", "TESTMODEL"]

        for model_name in model_names:
            response = admin_client.get(f"/api/2.0/mlflow/permissions/registered-models/{model_name}/users")

            # Should handle different cases (behavior may vary based on implementation)
            assert response.status_code == 200



import logging
import os
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urljoin

import httpx
import mlflow
from mlflow.entities import Feedback
from mlflow.genai.scorers import scorer as scorer_decorator
from mlflow.tracking import MlflowClient
from playwright.sync_api import Cookie

# Configure logging
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

from playwright.sync_api import sync_playwright, Page
from urllib.parse import quote


def bake_cookies(cookies: List[Cookie]) -> httpx.Cookies:
    ready_cookies = httpx.Cookies()
    for cookie in cookies:
        if "value" in cookie:
            name = cookie.get("name")
            if name is not None:
                value = cookie.get("value")
                if value is not None:
                    ready_cookies.set(name, value)
    return ready_cookies


def create_experiment(
    experiment_name: str,
    cookies,
    url: str = "http://localhost:8080/",
    create_api: str = "ajax-api/2.0/mlflow/experiments/create",
    get_api: str = "ajax-api/2.0/mlflow/experiments/get-by-name?experiment_name=",
) -> bool:
    respone = httpx.get(url=urljoin(url, get_api) + quote(experiment_name), cookies=cookies)
    if respone.status_code == 200:
        print(f"Experiment {experiment_name} already exists")
        return True
    if respone.status_code == 404:
        respone = httpx.post(url=urljoin(url, create_api), json={"name": experiment_name}, cookies=cookies)
        if respone.status_code == 200:
            print(f"Experiment {experiment_name} created")
            return True
        else:
            print(f"Experiment {experiment_name} not created")
            return False
    return False


def _build_seed_scorers() -> list[Any]:
    """Return a small set of demo scorers used when seeding test data."""

    @scorer_decorator
    def response_length(outputs: dict[str, Any] | None = None) -> Feedback:
        text = (outputs or {}).get("response", "")
        length = len(text)
        return Feedback(value=length, rationale=f"Response length = {length}", metadata={"chars": length})

    @scorer_decorator
    def contains_hello(outputs: dict[str, Any] | None = None) -> Feedback:
        text = (outputs or {}).get("response", "").lower()
        found = "hello" in text
        return Feedback(value=found, rationale=f"'hello' present: {found}", metadata={})

    return [response_length, contains_hello]


def create_model(
    model_name: str,
    cookies,
    url: str = "http://localhost:8080/",
    create_api: str = "ajax-api/2.0/mlflow/registered-models/create",
    get_api: str = "ajax-api/2.0/mlflow/registered-models/get?name=",
) -> bool:
    respone = httpx.get(url=urljoin(url, get_api) + quote(model_name), cookies=cookies)
    print(respone.status_code)
    if respone.status_code == 404:
        respone = httpx.post(url=urljoin(url, create_api), json={"name": model_name}, cookies=cookies)
        if respone.status_code == 200:
            print(f"Model {model_name} created")
            return True
        else:
            print(f"Model {model_name} not created")
            return False
    print(f"Model {model_name} already exists")
    return True


def create_prompt(
    prompt_name: str,
    prompt_text: str,
    cookies,
    url: str = "http://localhost:8080/",
    create_api: str = "ajax-api/2.0/mlflow/registered-models/create",
    create_version_api: str = "ajax-api/2.0/mlflow/model-versions/create",
    get_api: str = "ajax-api/2.0/mlflow/registered-models/get?name=",
    commit_message: str = "commit message here",
    source: str = "dummy-source",
) -> bool:
    respone = httpx.get(url=urljoin(url, get_api) + prompt_name, cookies=cookies)
    print(respone.status_code)
    if respone.status_code == 404:
        respone = httpx.post(
            url=urljoin(url, create_api), json={"name": prompt_name, "tags": [{"key": "mlflow.prompt.is_prompt", "value": "true"}]}, cookies=cookies
        )
        if respone.status_code == 200:
            print(f"Prompt {prompt_name} created")
            # Now create the version with the prompt text
            respone = httpx.post(
                url=urljoin(url, create_version_api),
                json={
                    "name": prompt_name,
                    "description": commit_message,
                    "source": source,
                    "tags": [{"key": "mlflow.prompt.is_prompt", "value": "true"}, {"key": "mlflow.prompt.text", "value": prompt_text}],
                },
                cookies=cookies,
            )
            if respone.status_code == 200:
                print(f"Prompt version for {prompt_name} created")
                return True
            else:
                print(f"Prompt version for {prompt_name} not created")
                return False
        else:
            print(f"Prompt {prompt_name} not created")
            return False
    print(f"Prompt {prompt_name} already exists")
    return True


def register_sample_scorers(
    experiment_name: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
    register_api: str = "ajax-api/2.0/mlflow/scorers/register",
) -> Tuple[bool, bool]:
    """
    Register a couple of demo scorers for an experiment so E2E tests have scorer data.

    Returns (success, missing_endpoint). missing_endpoint is True when the registration
    endpoint responds with 404, allowing callers to skip when unsupported.
    """

    experiment_id = get_experiment_id(experiment_name, cookies, url=url)
    if not experiment_id:
        logger.warning("Could not resolve experiment id for %s", experiment_name)
        return False, False

    success = True
    missing_endpoint = False
    for scorer_obj in _build_seed_scorers():
        payload = {
            "experiment_id": experiment_id,
            "name": scorer_obj.name,
            "serialized_scorer": scorer_obj.model_dump_json(),
        }
        response = httpx.post(url=urljoin(url, register_api), json=payload, cookies=cookies)
        if response.status_code == 404:
            logger.warning(
                "Scorer registration endpoint missing for %s: %s %s",
                experiment_name,
                response.status_code,
                response.text,
            )
            missing_endpoint = True
            break
        if response.status_code not in (200, 409):
            logger.warning(
                "Failed to register scorer %s for %s: %s %s",
                scorer_obj.name,
                experiment_name,
                response.status_code,
                response.text,
            )
            success = False
    return success, missing_endpoint


def create_service_account(
    username: str,
    display_name: str,
    cookies: httpx.Cookies,
    base_url: str,
    users_api: str = "api/2.0/mlflow/users",
) -> tuple[bool, str]:
    """Create a service account via the admin session cookies."""

    payload = {
        "username": username,
        "display_name": display_name,
        "is_admin": False,
        "is_service_account": True,
    }
    response = httpx.post(url=urljoin(base_url, users_api), json=payload, cookies=cookies, timeout=10.0)
    if response.status_code in (200, 201):
        return True, response.json().get("message", "created") if response.headers.get("content-type", "").startswith("application/json") else "created"

    return False, f"{response.status_code}: {response.text}"


def create_access_token_for_user(
    username: str,
    cookies: httpx.Cookies,
    base_url: str,
    access_token_api: str = "api/2.0/mlflow/users/access-token",
) -> tuple[bool, str]:
    """Request an access token for the target user using admin cookies."""

    response = httpx.patch(url=urljoin(base_url, access_token_api), json={"username": username}, cookies=cookies, timeout=10.0)
    if response.status_code == 404:
        return False, "access-token endpoint unavailable"

    if response.status_code not in (200, 201):
        return False, f"{response.status_code}: {response.text}"

    try:
        payload = response.json()
    except Exception as exc:  # pragma: no cover - defensive
        return False, f"Invalid JSON response: {exc}"

    token = payload.get("token") if isinstance(payload, dict) else None
    if not token:
        return False, f"Token missing in response: {payload}"

    return True, token


def _normalize_metric_value(value: Any) -> float:
    """Normalize feedback values to floats for MLflow metrics."""

    if isinstance(value, bool):
        return float(int(value))
    if isinstance(value, (int, float)):
        return float(value)
    return float(len(str(value)))


def seed_scorers_with_tracking_token(
    experiment_name: str,
    token: str,
    base_url: str,
    response_text: str = "Hello from integration",
    username: str | None = None,
) -> tuple[str, Dict[str, float]]:
    """Seed scorer metrics by running an MLflow job authenticated via access token.

    Uses Basic auth: username + token as password.
    """
    env_backup = {
        "MLFLOW_TRACKING_USERNAME": os.environ.get("MLFLOW_TRACKING_USERNAME"),
        "MLFLOW_TRACKING_PASSWORD": os.environ.get("MLFLOW_TRACKING_PASSWORD"),
        "MLFLOW_TRACKING_URI": os.environ.get("MLFLOW_TRACKING_URI"),
    }

    if username:
        os.environ["MLFLOW_TRACKING_USERNAME"] = username
        os.environ["MLFLOW_TRACKING_PASSWORD"] = token
    os.environ["MLFLOW_TRACKING_URI"] = base_url.rstrip("/")

    try:
        mlflow.set_tracking_uri(os.environ["MLFLOW_TRACKING_URI"])
        mlflow.set_experiment(experiment_name)

        with mlflow.start_run(run_name=f"seed-scorers-{experiment_name}") as run:
            for scorer_obj in _build_seed_scorers():
                feedback = scorer_obj({"response": response_text})
                value = _normalize_metric_value(feedback.value)
                mlflow.log_metric(f"scorer.{scorer_obj.name}", value)

            run_id = run.info.run_id

        client = MlflowClient()
        metrics = client.get_run(run_id).data.metrics
        return run_id, metrics
    finally:
        if env_backup["MLFLOW_TRACKING_USERNAME"] is not None:
            os.environ["MLFLOW_TRACKING_USERNAME"] = env_backup["MLFLOW_TRACKING_USERNAME"]
        else:
            os.environ.pop("MLFLOW_TRACKING_USERNAME", None)

        if env_backup["MLFLOW_TRACKING_PASSWORD"] is not None:
            os.environ["MLFLOW_TRACKING_PASSWORD"] = env_backup["MLFLOW_TRACKING_PASSWORD"]
        else:
            os.environ.pop("MLFLOW_TRACKING_PASSWORD", None)

        if env_backup["MLFLOW_TRACKING_URI"] is not None:
            os.environ["MLFLOW_TRACKING_URI"] = env_backup["MLFLOW_TRACKING_URI"]
        else:
            os.environ.pop("MLFLOW_TRACKING_URI", None)


def set_experiment_permission(
    experiment_id: str,
    user_email: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set user-level permission for an experiment.

    Returns (success, message). Uses PATCH to update or POST to create.
    """
    api = f"api/2.0/mlflow/permissions/users/{quote(user_email)}/experiments/{quote(experiment_id)}"
    full_url = urljoin(url, api)

    # Try PATCH first (update existing)
    resp = httpx.patch(url=full_url, json={"permission": permission}, cookies=cookies, timeout=10.0)
    if resp.status_code == 200:
        return True, "updated"

    # Try POST (create new)
    if resp.status_code == 404:
        resp = httpx.post(url=full_url, json={"permission": permission}, cookies=cookies, timeout=10.0)
        if resp.status_code in (200, 201):
            return True, "created"

    return False, f"{resp.status_code}: {resp.text}"


def set_model_permission(
    model_name: str,
    user_email: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set user-level permission for a registered model.

    Returns (success, message). Uses PATCH to update or POST to create.
    """
    api = f"api/2.0/mlflow/permissions/users/{quote(user_email)}/registered-models/{quote(model_name)}"
    full_url = urljoin(url, api)

    # Try PATCH first (update existing)
    resp = httpx.patch(url=full_url, json={"permission": permission}, cookies=cookies, timeout=10.0)
    if resp.status_code == 200:
        return True, "updated"

    # Try POST (create new)
    if resp.status_code == 404:
        resp = httpx.post(url=full_url, json={"permission": permission}, cookies=cookies, timeout=10.0)
        if resp.status_code in (200, 201):
            return True, "created"

    return False, f"{resp.status_code}: {resp.text}"


def set_prompt_permission(
    prompt_name: str,
    user_email: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set user-level permission for a prompt.

    Returns (success, message). Prompts use the same registered-models endpoint.
    """
    return set_model_permission(prompt_name, user_email, permission, cookies, url)


def get_experiment_id(experiment_name: str, cookies: httpx.Cookies, url: str = "http://localhost:8080/", api: str = "api/2.0/mlflow/experiments") -> str:
    """Resolve experiment_id using the get-by-name API (more reliable across deployments)."""

    get_by_name_api = "ajax-api/2.0/mlflow/experiments/get-by-name?experiment_name="
    response = httpx.get(url=urljoin(url, get_by_name_api) + quote(experiment_name), cookies=cookies)
    if response.status_code != 200:
        return ""

    payload = response.json()
    experiment = payload.get("experiment", payload)
    return str(experiment.get("experiment_id") or experiment.get("experimentId") or experiment.get("id") or "")


def set_group_experiment_permission(
    experiment_name: str,
    group_name: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set group-level permission for an experiment.

    Returns (success, message). Success is False when the experiment is not found or the
    endpoint is unavailable; callers can choose to skip on 404s.

    Uses the new RESTful API:
    - POST api/2.0/mlflow/permissions/groups/{group_name}/experiments/{experiment_id} to create
    - DELETE api/2.0/mlflow/permissions/groups/{group_name}/experiments/{experiment_id} to delete
    """

    experiment_id = get_experiment_id(experiment_name, cookies, url=url)
    if not experiment_id:
        return False, f"Experiment {experiment_name} not found"

    # New RESTful API path
    api_url = urljoin(url, f"api/2.0/mlflow/permissions/groups/{quote(group_name)}/experiments/{quote(experiment_id)}")

    # Delete existing permission first (ignore errors - permission may not exist)
    # Note: Server returns 500 when permission doesn't exist, not 404
    httpx.request(
        method="DELETE",
        url=api_url,
        cookies=cookies,
        timeout=10.0,
    )

    # Create new permission with POST
    set_response = httpx.post(
        url=api_url,
        json={"permission": permission},
        cookies=cookies,
        timeout=10.0,
    )
    if set_response.status_code in (200, 201):
        return True, "ok"

    return False, f"Set failed ({set_response.status_code}): {set_response.text}"
    # http://localhost:8080/api/2.0/mlflow/groups/mlflow-admin/experiments/delete
    # {"experiment_id":"537270198867596775"}
    # set new permissions
    # http://localhost:8080/api/2.0/mlflow/groups/mlflow-admin/experiments/create
    # POST
    # {"experiment_id":"537270198867596775","permission":"READ"}
    # get all experiments
    # http://localhost:8080/api/2.0/mlflow/experiments
    # [
    #     {
    #         "id": "0",
    #         "name": "Default",
    #         "tags": {}
    #     }
    # ]
    # pass


def set_group_model_permission(
    model_name: str,
    group_name: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set group-level permission for a registered model.

    Returns (success, message).

    Uses the new RESTful API:
    - POST api/2.0/mlflow/permissions/groups/{group_name}/registered-models/{name} to create
    - DELETE api/2.0/mlflow/permissions/groups/{group_name}/registered-models/{name} to delete
    """
    # New RESTful API path
    api_url = urljoin(url, f"api/2.0/mlflow/permissions/groups/{quote(group_name)}/registered-models/{quote(model_name)}")

    # Delete existing permission first (ignore errors - permission may not exist)
    # Note: Server returns 500 when permission doesn't exist, not 404
    httpx.request(
        method="DELETE",
        url=api_url,
        cookies=cookies,
        timeout=10.0,
    )

    # Create new permission with POST
    set_response = httpx.post(
        url=api_url,
        json={"permission": permission},
        cookies=cookies,
        timeout=10.0,
    )
    if set_response.status_code in (200, 201):
        return True, "ok"

    return False, f"Set failed ({set_response.status_code}): {set_response.text}"


def set_group_prompt_permission(
    prompt_name: str,
    group_name: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set group-level permission for a prompt.

    Returns (success, message). Prompts use registered-models endpoint.
    """
    return set_group_model_permission(prompt_name, group_name, permission, cookies, url)


def set_regexp_experiment_permission(
    pattern: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
    api: str = "api/2.0/mlflow/permissions/experiments/regex",
) -> tuple[bool, str]:
    """Set regex-based permission for experiments.

    Returns (success, message).
    """
    full_url = urljoin(url, api)
    resp = httpx.post(
        url=full_url,
        json={"regex": pattern, "permission": permission},
        cookies=cookies,
        timeout=10.0,
    )
    if resp.status_code in (200, 201):
        return True, "created"
    if resp.status_code == 409:
        return True, "already exists"
    return False, f"{resp.status_code}: {resp.text}"


def set_regexp_model_permission(
    pattern: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
    api: str = "api/2.0/mlflow/permissions/registered-models/regex",
) -> tuple[bool, str]:
    """Set regex-based permission for registered models.

    Returns (success, message).
    """
    full_url = urljoin(url, api)
    resp = httpx.post(
        url=full_url,
        json={"regex": pattern, "permission": permission},
        cookies=cookies,
        timeout=10.0,
    )
    if resp.status_code in (200, 201):
        return True, "created"
    if resp.status_code == 409:
        return True, "already exists"
    return False, f"{resp.status_code}: {resp.text}"


def set_regexp_prompt_permission(
    pattern: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set regex-based permission for prompts.

    Returns (success, message). Uses registered-models regex endpoint.
    """
    return set_regexp_model_permission(pattern, permission, cookies, url)


def set_regexp_group_experiment_permission(
    pattern: str,
    group_name: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
    api: str = "api/2.0/mlflow/permissions/groups/{group_name}/experiments/regex",
) -> tuple[bool, str]:
    """Set group-regex permission for experiments.

    Returns (success, message).
    """
    full_url = urljoin(url, api.format(group_name=quote(group_name)))
    resp = httpx.post(
        url=full_url,
        json={"regex": pattern, "permission": permission},
        cookies=cookies,
        timeout=10.0,
    )
    if resp.status_code in (200, 201):
        return True, "created"
    if resp.status_code == 409:
        return True, "already exists"
    return False, f"{resp.status_code}: {resp.text}"


def set_regexp_group_model_permission(
    pattern: str,
    group_name: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
    api: str = "api/2.0/mlflow/permissions/groups/{group_name}/registered-models/regex",
) -> tuple[bool, str]:
    """Set group-regex permission for registered models.

    Returns (success, message).
    """
    full_url = urljoin(url, api.format(group_name=quote(group_name)))
    resp = httpx.post(
        url=full_url,
        json={"regex": pattern, "permission": permission},
        cookies=cookies,
        timeout=10.0,
    )
    if resp.status_code in (200, 201):
        return True, "created"
    if resp.status_code == 409:
        return True, "already exists"
    return False, f"{resp.status_code}: {resp.text}"


def set_regexp_group_prompt_permission(
    pattern: str,
    group_name: str,
    permission: str,
    cookies: httpx.Cookies,
    url: str = "http://localhost:8080/",
) -> tuple[bool, str]:
    """Set group-regex permission for prompts.

    Returns (success, message). Uses registered-models regex endpoint.
    """
    return set_regexp_group_model_permission(pattern, group_name, permission, cookies, url)


def user_login(page: Page, username: str, url: str = "http://localhost:8080/") -> httpx.Cookies:
    """Login user using Playwright and httpx cookies.
    This function navigates to the login page, fills in the username, and returns the cookies."""
    try:
        print(f"🔍 Navigating to {url}")
        page.goto(url)
        page.wait_for_load_state("networkidle")

        # print(f"📸 Taking screenshot of initial page")
        # page.screenshot(path=f"debug-initial-{username}.png")

        print(f"🔍 Looking for login button...")
        # Look for any button containing "login" or "Login"
        login_buttons = page.locator("button").filter(has_text="Login")
        if login_buttons.count() == 0:
            print("❌ No login button found, checking page content...")
            print(f"Page title: {page.title()}")
            print(f"Page URL: {page.url}")

            # Check if we're already on a login page
            if "login" in page.url.lower() or "oauth" in page.url.lower():
                print("✅ Already on login/oauth page")
            else:
                print("❌ Not on login page and no login button found")
                raise Exception("No login button found and not on login page")
        else:
            print(f"✅ Found {login_buttons.count()} login button(s)")
            # Click the first login button
            login_buttons.first.click()
            page.wait_for_load_state("networkidle")
            # print(f"📸 Taking screenshot after login button click")
            # page.screenshot(path=f"debug-after-login-click-{username}.png")

        print(f"🔍 Looking for username field...")
        # Look for username/sub field
        username_field = page.locator("input[name='sub']").or_(page.locator("input[type='text']")).first
        if username_field.is_visible():
            print(f"✅ Found username field, filling with '{username}'")
            username_field.fill(username)
        else:
            print("❌ Username field not found")
            raise Exception("Username field not found")

        # print(f"📸 Taking screenshot after filling username")
        # page.screenshot(path=f"debug-after-username-{username}.png")

        print(f"🔍 Looking for authorize/submit button...")
        # Look for authorize button
        auth_buttons = page.locator("button").filter(has_text="Authorize").or_(page.locator("button[type='submit']"))
        if auth_buttons.count() > 0:
            print(f"✅ Found authorize button, clicking...")
            auth_buttons.first.click()
            page.wait_for_load_state("networkidle")
        else:
            print("❌ Authorize button not found")
            raise Exception("Authorize button not found")

        # print(f"📸 Taking screenshot after authorize")
        # page.screenshot(path=f"debug-after-authorize-{username}.png")

        print(f"✅ Login flow completed for {username}")
        print(f"Final URL: {page.url}")

        cookies = page.context.cookies()
        print(f"🍪 Got {len(cookies)} cookies")
        for cookie in cookies:
            print(f"  - {cookie['name']}: {cookie['value'][:20]}...")  # type: ignore

        return bake_cookies(cookies)

    except Exception as e:
        logger.error(f"Failed to login user {username}: {e}")
        page.screenshot(path=f"debug-error-{username}.png")
        raise
